from tnreason.logic import coordinate_calculus as cc
import numpy as np


class Layer:
    def __init__(self, coresDict={}, signsDict={}):
        self.signsDict = signsDict
        self.coresDict = coresDict

        self.colors = []
        for key in self.coresDict:
            self.colors = self.colors + self.coresDict[key].colors

    def compute_size(self):
        self.size = 0
        for key in self.coresDict:
            self.size += np.prod(self.coresDict[key].values.shape)
        return self.size

    def reduce_color(self, color):
        for key in self.coresDict:
            if color in self.coresDict[key].colors:
                self.coresDict[key] = self.coresDict[key].reduce_color(color)
        return self

    def switch_sign(self):
        for key in self.signsDict:
            self.signsDict[key] = -1 * self.signsDict[key]


def contract(layer0, layer1):
    contractedCoresDict = {}
    contractedSignsDict = {}
    for key0 in layer0.coresDict:
        for key1 in layer1.coresDict:
            contractedSignsDict[key0 + "_" + key1] = layer0.signsDict[key0] * layer1.signsDict[key1]
            contractedCoresDict[key0 + "_" + key1] = layer0.coresDict[key0].compute_and(layer1.coresDict[key1])
    return Layer(contractedCoresDict, contractedSignsDict)


if __name__ == "__main__":
    l = Layer({"a": cc.CoordinateCore(core_values=np.array([2]), core_colors=["b1"])}, {"a": 1})

    print(l.compute_size())
    print(l.reduce_color("b1"))
    print(contract(l, l).coresDict)



from tnreason.model import tensor_model as tm
from tnreason.model import formula_tensors as ft

from tnreason.contraction import core_contractor as coc


class TensorKB:
    def __init__(self, formulaList):
        self.formulaList = formulaList
        self.formulaTensors = tm.TensorRepresentation(
            {"f" + str(i): [formula, 1] for i, formula in enumerate(formulaList)}, headType="truthEvaluation")

        print("Initialized a Knowledge Base with the formulas: \n  {}".format(
            "\n  ".join([str(f) for f in self.formulaList])))

    def tell(self, formula):
        assert self.formulaTensors.headType == "truthEvaluation"

        answer = self.ask(formula)
        if answer == "entailed":
            print("{} is redundant to the Knowledge Base and has not been added.".format(formula))
            return "not added"
        elif answer == "contradicting":
            print("{} would make the Knowledge Base inconsistent and has not been added.".format(formula))
            return "not added"
        else:
            self.formulaList.append(formula)
            self.formulaTensors.add_expression(formula, formulaKey="f" + str(len(self.formulaList)))

            print("{} has been added to the Knowledge Base.".format(formula))
            return "added"
    def ask(self, formula):
        if coc.CoreContractor({**self.formulaTensors.get_cores(headType="truthEvaluation"),
                               **ft.FormulaTensor(expression=["not", formula],
                                                  headType="truthEvaluation").get_cores()
                               }).contract().values == 0:
            print("{} is entailed by the Knowledge Base.".format(formula))
            return "entailed"
        elif coc.CoreContractor({**self.formulaTensors.get_cores(headType="truthEvaluation"),
                                 **ft.FormulaTensor(expression=formula,
                                                    headType="truthEvaluation").get_cores()
                                 }).contract().values == 0:
            print("{} is contradicted by the Knowledge Base.".format(formula))
            return "contradicting"
        else:
            print("{} is neither entailed nor contradicted by the Knowledge Base.".format(formula))
            return "contingent"



import numpy as np
# import tnreason.logic.coordinate_calculus as cc
# import tnreason.logic.basis_calculus as bc

## SHIFTED TO THESE MODULES
import tnreason.logic.expression_utils as eu
import tnreason.contraction.expression_evaluation as ee


## Replaced by tnreason.contraction.expression_evaluation -> ExpressionEvaluator(expression, atomDict).evaluate()
def calculate_core(atom_dict, expression):
    return ee.ExpressionEvaluator(expression, atomDict=atom_dict).evaluate()
    # if type(expression) == str:
    #     return atom_dict[expression]
    # elif expression[0] == "not":
    #     return calculate_core(atom_dict, expression[1]).negate()
    # elif expression[1] == "and":
    #     return calculate_core(atom_dict, expression[0]).compute_and(
    #         calculate_core(atom_dict, expression[2]))
    # else:
    #     raise ValueError("Expression {} not understood.".format(expression))


## To be replaced by tnreason.contraction.expression_evaluation -> ExpressionEvaluator(expression).evaluate_on_sampleDf()
def evaluate_expression_on_sampleDf(sampleDf, expression):
    return ee.ExpressionEvaluator(expression).evaluate_on_sampleDf(sampleDf)
    # variables = eu.get_variables(expression)
    # atomDict = {}
    # for variable in variables:
    #     if variable == "Thing":
    #         values = np.ones(sampleDf.shape[0])
    #     elif variable == "Nothing":
    #         values = np.zeros(sampleDf.shape[0])
    #     else:
    #         values = sampleDf[variable].astype("int64").values
    #     atomDict[variable] = cc.CoordinateCore(values, ["j"], variable)
    # return calculate_core(atomDict, expression)


###

def evaluate_expression_on_factDf(factDf, individualsDict, expression):
    variables = eu.get_variables(expression)
    atomDict = generate_atomDict(factDf, individualsDict, variables)
    return calculate_core(atomDict, expression)


def generate_atomDict(factDf, individualsDict, atoms):
    atomDict = {}
    for atomKey in atoms:
        if "," in atomKey:
            relationKey = atomKey.split("(")[0]
            indKey1, indKey2 = atomKey.split("(")[1][:-1].split(",")
            relValues = generate_relation_values(factDf, individualsDict[indKey1], individualsDict[indKey2],
                                                 relationKey)
            atomDict[atomKey] = cc.CoordinateCore(relValues, [indKey1, indKey2], atomKey)
        else:
            classKey = atomKey.split("(")[0]
            indKey = atomKey.split("(")[1][:-1]
            classValues = generate_class_values(factDf, individualsDict[indKey], classKey)
            atomDict[atomKey] = cc.CoordinateCore(classValues, [indKey], atomKey)
    return atomDict


def generate_class_values(factDf, individuals, classKey):
    relevantInds = factDf[factDf["predicate"].isin(["http://www.w3.org/1999/02/22-rdf-syntax-ns#type", "rdf:type"])]
    relevantInds = relevantInds[relevantInds["object"] == classKey]
    outvalues = np.zeros((len(individuals)))
    for i, row in relevantInds.iterrows():
        subPos = np.argwhere(individuals == row["subject"])
        outvalues[subPos] = 1
    return outvalues


def generate_relation_values(factDf, individuals1, individuals2, relationKey):
    relevantPairs = factDf[factDf["predicate"] == relationKey]
    outValues = np.zeros((len(individuals1), len(individuals2)))
    for i, row in relevantPairs.iterrows():
        subPos = np.argwhere(individuals1 == row["subject"])
        obPos = np.argwhere(individuals2 == row["object"])
        outValues[subPos, obPos] = 1
    return outValues


def calculate_expressionCore(expression):
    variables = np.unique(eu.get_variables(expression))
    atom_dict = {}
    for variable in variables:
        atom_dict[variable] = bc.BasisCore(np.eye(2), [variable, "head"], headcolor="head", name=variable)
    return calculate_core(atom_dict, expression).calculate_truth().reduce_identical_colors().to_coordinate()


if __name__ == "__main__":
    from tnreason.logic import basis_calculus as bc, coordinate_calculus as cc

    core0_values = np.random.normal(size=(100, 10, 5))
    core0_colors = ["a", "b", "c"]
    core1_values = np.random.normal(size=(10, 100, 3))
    core1_colors = ["b", "a", "d"]

    core0 = cc.CoordinateCore(core0_values, core0_colors, "Sledz")
    core1 = cc.CoordinateCore(core1_values, core1_colors, "Jaszczur")
    core2 = cc.CoordinateCore(core1_values, core1_colors, "Sokol")

    atom_dict = {"a": core0,
                 "b": core1,
                 "c": core2}
    # example_expression = ["a","and",["not",["b","and","b"]]]
    example_expression = [[["not", "a"], "and", ["not", "b"]], "and", "b"]
    core = calculate_core(atom_dict, example_expression)
    print(core.name)
    print(core.values.shape)
    # example_expression = [["a", "and", "b"], "and", ["not", "c"]]

    core0 = bc.BasisCore(np.ones(2), ["Sledz"])
    core1 = bc.BasisCore(np.ones(2), ["Jaszczur"])
    core2 = bc.BasisCore(np.ones(2), ["Sokol"])

    atom_dict = {"a": core0,
                 "b": core1,
                 "c": core2}
    # example_expression = ["a","and",["not",["b","and","b"]]]
    example_expression = [[["not", "a"], "and", ["not", "b"]], "and", "b"]
    core = calculate_core(atom_dict, example_expression)
    print(core.name)
    print(core.values.shape)



from tnreason.logic import expression_utils as eu

def count_satisfaction(expression):
    conjunctionList = compute_conjunctionList(expression)
    count = 0
    for entry in conjunctionList:
        count += 2**len(entry[0])*entry[2]
    return count

def compute_conjunctionList(expression):
    if type(expression) == str:
        return [[[],[expression],1]]
    if expression[0] == "not":
        variables = combine_unique(eu.get_variables(expression),[])
        conjunctionList1 = compute_conjunctionList(expression[1])
        conjunctionList = [[variables,[],1]]
        for entry in conjunctionList1:
            conjunctionList.append([entry[0],entry[1],-1*entry[2]])
        return conjunctionList
    elif expression[1] == "and":
        conList0 =  compute_conjunctionList(expression[0])
        conList2 =  compute_conjunctionList(expression[2])

        conjunctionList = []
        for entry0 in conList0:
            for entry2 in conList2:
                conjunctionList.append(merge_entries(entry0,entry2))
        return conjunctionList
    else:
        raise ValueError("Expression {} not understood.".format(expression))

def merge_entries(entry0, entry2):
    touchedVariables = entry0[1]+entry2[1]
    preuntouchedVariables = combine_unique(entry0[0],entry2[0])
    untouchedVariables = []
    for variable in preuntouchedVariables:
        if variable not in touchedVariables:
            untouchedVariables.append(variable)
    return [untouchedVariables, touchedVariables, entry0[2]*entry2[2]]

def combine_unique(list1,list2):
    list = []
    for element in list1+list2:
        if element not in list:
            list.append(element)
    return list

if __name__ == "__main__":
    assert count_satisfaction("sledz")==1, "Satisfaction Counter wrong"
    assert count_satisfaction(["sikorka","and",["not",["jaszczur","and","sledz"]]])==3, "Satisfaction Counter wrong"
    assert count_satisfaction(["not",["test","and","sledz"]])==3, "Satisfaction Counter wrong"
    assert count_satisfaction(["not",["losos","and",["sledz","and","sikorka"]]])==7, "Satisfaction Counter wrong"
    assert count_satisfaction(["not",["losos","and",["losos","and","sikorka"]]])==3, "Satisfaction Counter wrong"

    #print(compute_conjunctionList(["jaszczur","and",["not",["jaszczur","and","sledz"]]]))
    print(count_satisfaction(["sledz","and",["not","sledz"]]))
    assert count_satisfaction(["jaszczur","and",["not",["jaszczur","and","sledz"]]])==1, "Satisfaction Counter wrong"


import networkx as nx
import matplotlib.pyplot as plt

def generate_expression_graph(expression):
    if type(expression) == str:
        atomNodes = [expression]
        expressionNodes = []
        edgesAnd = []
        edgesNot = []
    elif expression[0] == "not":
        atomNodes, expressionNodes, edgesAnd, edgesNot = generate_expression_graph(expression[1])
        edgesNot.append([expression[1], expression])
    elif expression[1] == "and":
        atomNodes0, expressionNodes0, edgesAnd0, edgesNot0 = generate_expression_graph(expression[0])
        atomNodes2, expressionNodes2, edgesAnd2, edgesNot2 = generate_expression_graph(expression[2])
        atomNodes = atomNodes0 + atomNodes2
        expressionNodes = expressionNodes0 + expressionNodes2
        edgesAnd = edgesAnd0 + edgesAnd2
        edgesNot = edgesNot0 + edgesNot2
        edgesAnd.append([expression[0], expression])
        edgesAnd.append([expression[2], expression])

    expressionNodes.append(expression)
    return atomNodes, expressionNodes, edgesAnd, edgesNot

def visualize_expression_graph(expression,fontsize=16):
    atomNodes, expressionNodes, edgesAnd, edgesNot = generate_expression_graph(expression)

    graph = nx.DiGraph()
    graph.add_nodes_from([str(node) for node in expressionNodes])
    graph.add_edges_from([[str(edge[0]), str(edge[1])] for edge in edgesAnd + edgesNot])

    pos = nx.spring_layout(graph)

    labels = { atom : atom for atom in atomNodes}
    nx.draw_networkx_labels(graph, pos, labels, font_size=fontsize)

    nx.draw_networkx_nodes(graph, pos,
                           nodelist=atomNodes,
                           node_color='b',
                           node_size=3000,
                           alpha=0.2)
    nx.draw_networkx_nodes(graph, pos,
                           nodelist=[str(node) for node in expressionNodes if node not in atomNodes],
                           node_color='b',
                           node_size=1000,
                           alpha=0.1)
    nx.draw_networkx_nodes(graph, pos,
                           nodelist=[str(expression)],
                           node_color="black",
                           node_size=1000,
                           alpha=0.4)

    nx.draw_networkx_edges(graph, pos,
                           edgelist=[[str(edge[0]), str(edge[1])] for edge in edgesAnd],
                           arrowstyle="-|>",
                           width=2, alpha=1, edge_color='b')
    nx.draw_networkx_edges(graph, pos,
                           edgelist=[[str(edge[0]), str(edge[1])] for edge in edgesNot],
                           width=2, alpha=1, edge_color='r')
    plt.show()


if __name__ == "__main__":
   visualize_expression_graph(["R2(x,z)", "and", ["C1(x)", "and", ["not", "R1(y,x)"]]],fontsize=10)




from matplotlib import pyplot as plt

from tnreason.contraction import contraction_optimization as co
from tnreason.contraction import bc_contraction_generation as cg
from tnreason.contraction import contraction_visualization as cv


class CoreContractor:
    """
    coreDict: Dictionary of CoordinateCores
    coreList: Order of coreDict keys for contraction
    instructionList: list of contraction instructions: either "and" with additional core or "reduce" with a color. First entry must be add to start with.
    """

    def __init__(self, coreDict={}, coreList=None, instructionList=None, openColors=[]):
        self.coreDict = coreDict
        self.coreList = coreList
        self.instructionList = instructionList
        self.openColors = openColors

    def generate_coreDict_from_formulaList(self, formulaList):
        for formula in formulaList:
            self.coreDict = {**self.coreDict, **cg.generate_factor_dict(formula)}

    def optimize_coreList(self, method="GreedyHeuristic"):
        # Generate the coreColorDict and colorDimDict for ContractionOptimizer
        coreColorDict = {}
        colorDimDict = {}
        for coreKey in self.coreDict:
            coreColorDict[coreKey] = self.coreDict[coreKey].colors.copy()
            for i, color in enumerate(self.coreDict[coreKey].colors):
                if color not in colorDimDict:
                    colorDimDict[color] = self.coreDict[coreKey].values.shape[i]
        if method == "GreedyHeuristic":
            optimizer = co.GreedyHeuristicOptimizer(coreColorDict, colorDimDict, globallyOpenColors=self.openColors)
            # Optimize coreList i.e. order of contraction
            optimizer.optimize()
        elif method == "Dijkstra":
            optimizer = co.DijkstraOptimizer(coreColorDict, colorDimDict, globallyOpenColors=self.openColors)
            optimizer.optimize()
        self.coreList = optimizer.coreList

    def create_instructionList_from_coreList(self, verbose=False):
        if self.coreList is None:
            self.coreList = list(self.coreDict.keys())
        # Find all colors
        colorList = []
        for key in self.coreDict:
            for color in self.coreDict[key].colors:
                if color not in colorList:
                    colorList.append(color)
        # Find cores after which color can be reduced
        self.coreList = list(self.coreList)
        self.coreList.reverse()
        reduceDict = {key: [] for key in self.coreDict}
        for color in colorList:
            if color not in self.openColors:
                for key in self.coreList:
                    if color in self.coreDict[key].colors:
                        reduceDict[key].append(color)
                        break
        self.coreList.reverse()
        # Create the instructionList
        self.instructionList = []
        for key in self.coreList:
            self.instructionList.append(["add", key])
            for color in reduceDict[key]:
                self.instructionList.append(["reduce", color])
        if verbose:
            print("The instructionList is {} and colorList {}.".format(self.instructionList,
                                                                       {key: self.coreDict[key].colors for key in
                                                                        self.coreDict}))

    def evaluate_sizes_instructionList(self, show=True):
        shapeList = [[]]
        colorList = [[]]
        for instruction in self.instructionList:
            shapes = shapeList[-1].copy()
            colors = colorList[-1].copy()
            if instruction[0] == "add":
                for i, color in enumerate(self.coreDict[instruction[1]].colors):
                    if color not in colors:
                        colors.append(color)
                        shapes.append(self.coreDict[instruction[1]].values.shape[i])
            elif instruction[0] == "reduce":
                popindex = colors.index(instruction[1])
                shapes.pop(popindex)
                colors.pop(popindex)
            shapeList.append(shapes)
            colorList.append(colors)
        sizeList = []
        for shapes in shapeList:
            sizeList.append(np.prod(shapes))

        if show:
            plt.title("Tensor Sizes During the Contraction", fontsize=15)
            plt.scatter(range(1, len(sizeList)), sizeList[1:], marker="+")
            plt.xticks(range(1, len(sizeList)), [str(ins) for ins in self.instructionList])
            plt.xlabel("Instructions")
            plt.ylabel("Number of Coordinates of the contracted")
            plt.ylim([0, 1.1 * max(sizeList)])
            plt.show()

        return sizeList, shapeList, colorList

    def visualize(self, title="Contraction Diagram"):
        cv.draw_contractionDiagram(self.coreDict, title=title)

    def contract(self, optimizationMethod=None, verbose=False):
        if optimizationMethod is None:
            if self.instructionList is None:
                self.create_instructionList_from_coreList()
        elif optimizationMethod == "GreedyHeuristic":
            self.optimize_coreList()
            self.create_instructionList_from_coreList()
        else:
            raise ValueError("Optimization Method {} not supported!".format(optimizationMethod))
        contracted = self.coreDict[self.instructionList[0][1]]
        for instruction in self.instructionList[1:]:
            if verbose:
                print("## Doing {} ##".format(instruction))
            if instruction[0] == "add":
                contracted = contracted.compute_and(self.coreDict[instruction[1]])
            elif instruction[0] == "reduce":
                contracted = contracted.reduce_color(instruction[1])
            elif instruction[0] == "exp":
                contracted = contracted.exponenentiate()
            else:
                raise ValueError("Instruction {} not understood.".format(instruction))
        if verbose and len(contracted.values.shape) > 0:
            print("Missing contraction colors are {}.".format(contracted.colors))
        return contracted

    ## Unused!
    def contract_color(self, color):
        affectedCores, affectedKeys = find_affected(self.coreDict, color)

        contracted = self.coreDict[affectedKeys[0]]
        for i, key in enumerate(affectedKeys[1:]):
            contracted = contracted.compute_and(self.coreDict[key])

        contracted.count_on_color(color)

        self.coreDict = {key: self.coreDict[key] for key in self.coreDict if key not in affectedKeys}
        self.coreDict["contracted_" + color] = contracted

        return contracted

    def numpy_einsum_contract(self):
        subscripts = self.create_contraction_subscripts()

        contractedValues = np.einsum(subscripts, *[self.coreDict[coreKey].values for coreKey in self.coreList])
        return cc.CoordinateCore(contractedValues, self.openColors)

    def create_contraction_subscripts(self):
        alphabet = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
                    'u', 'v', 'w', 'x', 'y', 'z']

        colorDict = {}
        i = 0
        for coreKey in self.coreDict:
            for color in self.coreDict[coreKey].colors:
                if color not in colorDict:
                    colorDict[color] = alphabet[i]
                    i += 1

        colorList = []
        for coreKey in self.coreList:
            colorList.append("".join([colorDict[color] for color in self.coreDict[coreKey].colors]))

        lhs = ",".join(colorList)
        rhs = "".join([colorDict[color] for color in self.openColors])

        return lhs + "->" + rhs


def find_affected(coreDict, color):
    affectedCores = []
    affectedKeys = []
    for key in coreDict:
        if color in coreDict[key].colors:
            affectedCores.append(coreDict[key])
            affectedKeys.append(key)
    return affectedCores, affectedKeys


## To replace Optimization Calculus
class NegationTolerantCoreContractor:
    ## In CoreDict
    def __init__(self, coreDict={}, coreList=None, instructionList=None, openColors=[]):
        self.coreDict = coreDict  ## Now each key gives a list of list [core, ignorecolors]
        self.coreList = coreList
        self.instructionList = instructionList
        self.openColors = openColors

    def contract(self):

        contractedList = self.coreDict[self.instructionList[0][1]]
        print(len(contractedList))
        for instruction in self.instructionList[1:]:
            if instruction[0] == "add":
                contractedList = compute_list_and(contractedList, self.coreDict[instruction[1]])
                #            elif instruction[0] == "reduce":
                #                contracted = contracted.reduce_color(instruction[1])
                print(instruction, len(contractedList))
        return contractedList


def compute_list_and(leftList, rightList):
    preResultList = []

    ## Add the non Constants
    for leftCore, leftIgnoreColors in leftList:
        for rightCore, rightIgnoreColors in rightList:
            if (not is_constant(leftIgnoreColors, rightCore.colors)) and (
                    not is_constant(leftIgnoreColors, rightCore.colors)):
                preResultList.append([leftCore.compute_and(rightCore), leftIgnoreColors + rightIgnoreColors])

    ## Add the constants from both sides
    for leftCore, leftIgnoreColors in leftList:
        firstRightCore, firstRightIgnoreColors = rightList[0]
        if is_constant(leftIgnoreColors, firstRightCore.colors):
            preResultList.append([leftCore, leftIgnoreColors])
    for rightCore, rightIgnoreColors in rightList:
        firstLeftCore, firstLeftIgnoreColors = leftList[0]
        if is_constant(rightIgnoreColors, firstLeftCore.colors):
            preResultList.append([rightCore, rightIgnoreColors])

    ## Sum the cores with same colors
    resultList = []
    while len(preResultList) > 0:
        core1, ignoreColors1 = preResultList.pop()
        for core2, ignoreColors2 in preResultList.copy():
            if color_equivalence(core1, core2):
                core1 = core1.sum_with(core2)
                preResultList.pop(preResultList.index([core2, ignoreColors2]))
        resultList.append([core1, ignoreColors1])
    return resultList


def is_constant(ignoreColors, testColors):
    for color in ignoreColors:
        if color in testColors:
            return True
    return False


def color_equivalence(core1, core2):
    for color in core1.colors:
        if color not in core2.colors:
            return False
    for color in core2.colors:
        if color not in core1.colors:
            return False
    return True


if __name__ == "__main__":
    import tnreason.logic.coordinate_calculus as cc
    import numpy as np


    def random_basis():
        vector = np.zeros((2))
        if np.random.binomial(n=1, p=0.5) == 1:
            vector[0] = 1
        else:
            vector[1] = 1
        return vector


    def calculate_random_basis_core(shape):
        shapeProduct = np.prod(shape)
        core = np.zeros([2, shapeProduct])
        for i in range(shapeProduct):
            core[:, i] = random_basis()
        return core.reshape([2] + shape)


    coordinateDict = {
        "a": cc.CoordinateCore(np.random.binomial(n=1, p=0.4, size=(10, 7, 5)), ["x", "y", "z"], name="a"),
        "b": cc.CoordinateCore(np.random.binomial(n=1, p=0.4, size=(10, 7, 5)), ["x", "q", "z"], name="b"),
    }

    weightDict = {
        "a": 2.134,
        "b": 1.72345
    }

    contractor = NegationTolerantCoreContractor(
        coreDict={
            "a": [
                [cc.CoordinateCore(np.random.binomial(n=1, p=0.4, size=(10, 7, 5)), ["x", "y", "z"], name="a"), ["y"]],
                [cc.CoordinateCore(np.random.binomial(n=1, p=0.4, size=(10, 7, 5)), ["x", "r", "z"], name="a"), ["y2"]]
            ],
            "b": [[cc.CoordinateCore(np.random.binomial(n=1, p=0.4, size=(10, 7, 5)), ["x", "y2", "z"], name="a2"),
                   ["y2"]],
                  [cc.CoordinateCore(np.random.binomial(n=1, p=0.4, size=(10, 7, 5)), ["x", "r", "z"], name="a"),
                   ["y2"]]
                  ],
        },
        instructionList=[["add", "a"], ["add", "b"], ["reduce", "x"], ["reduce", "y"]],
    )
    result = contractor.contract()
    print(result)

    exit()

    contractor = CoreContractor(coordinateDict,
                                None,
                                [["add", "a"], ["add", "b"], ["reduce", "x"], ["reduce", "y"]],
                                ["q"]
                                )

    contractor.optimize_coreList()
    contractor.create_instructionList_from_coreList()
    print(contractor.instructionList)  ## Hast to reduce colors only after last usage.

    contractorWithOpen = CoreContractor(coordinateDict, instructionList=None, coreList=None, openColors=["x"])
    contractorWithOpen.create_instructionList_from_coreList()
    print(contractorWithOpen.instructionList)  ## Has to be without openColor Reduction, i.e. "x".
    # contractorWithOpen.evaluate_sizes_instructionList(show=True)

    print(contractor.create_contraction_subscripts())
    print(contractor.numpy_einsum_contract().values)



from tnreason.logic import expression_utils as eu

import numpy as np


def refine_expression(expression, mode, newAtom):
    if mode == "not":
        return ["not", expression]
    if mode == "and":
        return [newAtom, "and", expression]


def add_leaf_atom(expression):
    atoms = eu.get_variables(expression)
    new_atom = "P" + str(len(atoms) + 1)
    return random_modification(expression, new_atom)


def random_modification(expression, atom_name="NewAtom"):
    if type(expression) == str:
        if np.random.rand() < 0.5:
            return refine_expression(expression, "not", atom_name)
        else:
            return refine_expression(expression, "and", atom_name)
    elif expression[0] == "not":
        return ["not", random_modification(expression[1], atom_name)]
    elif expression[1] == "and":
        if np.random.rand() < 0.5:
            left = random_modification(expression[0], atom_name)
            right = expression[2]
        else:
            left = expression[0]
            right = random_modification(expression[2], atom_name)
        return [left, "and", right]


if __name__ == "__main__":
    print(add_leaf_atom([["not", "sledz"], "and", "sikorka"]))



import numpy as np

def review_coreDict(coreDict):
    for coreKey in coreDict:
        print("## Investigating Core {} ##".format(coreKey))
        core = coreDict[coreKey]
        review_core(core)

def review_core(core):
    print("Squared Norm is {}".format(np.linalg.norm(core.values) ** 2))
    print("Shape is {}".format(core.values.shape))
    print("Colors are {}".format(core.colors))


from tnreason.model import sampling as sam
from tnreason.representation import sampledf_to_factdf as stof, sampledf_to_pairdf as stop

def generate_sampleDf(expressionDict, sampleNum, method="Gibbs10"):
    if method.startswith("Gibbs"):
        return sam.GibbsSampler(expressionDict).create_sampleDf(sampleNum, int(method.split("Gibbs")[1]))

def generate_factDf_and_pairDf(expressionDict,sampleNum,chainSize = 10, prefix ="tev"):
    sampleDf = generate_sampleDf(expressionDict, sampleNum, chainSize=chainSize)
    factDf = stof.sampleDf_to_factDf(sampleDf, prefix = prefix)
    pairDf = stop.sampleDf_to_pairDf(sampleDf, prefix = prefix)
    return factDf, pairDf




from tnreason.contraction import layers as lay
from tnreason.contraction import layered_contraction_generation as lcg

class LayeredCoreContractor:

    def __init__(self, layerDict={}, layerList=None, instructionList=None, openColors=[]):
        self.layerDict = layerDict
        self.layerList = layerList

        self.instructionList = instructionList
        self.openColors = openColors

    def contract(self):
        contracted = self.layerDict[self.instructionList[0][1]]
        for instruction in self.instructionList[1:]:
            if instruction[0] == "add":
                contracted = lay.contract(contracted, self.layerDict[instruction[1]])
            elif instruction[0] == "reduce":
                contracted = contracted.reduce_color(instruction[1])
        return contracted

    ## From analogous method in standard CoreContractor
    def create_instructionList_from_layerList(self):

        # Find all colors
        colorList = []
        for key in self.layerDict:
            for color in self.layerDict[key].colors:
                if color not in colorList and color not in self.openColors:
                    colorList.append(color)

        # Find cores after which color can be reduced
        self.layerList = list(self.layerList)
        self.layerList.reverse()
        reduceDict = {key: [] for key in self.layerDict}
        for color in colorList:
            if color not in self.openColors:
                for key in self.layerList:
                    if color in self.layerDict[key].colors:
                        reduceDict[key].append(color)
                        break
        self.layerList.reverse()

        # Create the instructionList
        self.instructionList = []
        for key in reduceDict:
            self.instructionList.append(["add", key])
            for color in reduceDict[key]:
                self.instructionList.append(["reduce", color])

if __name__ == "__main__":
    layDict = lcg.generate_skeletonLayerDict(["a","and",["not","b"]], {"a": 2, "b": 10}, 4)
    contractor = LayeredCoreContractor(layDict, list(layDict.keys()))
    contractor.create_instructionList_from_layerList()
    print(contractor.contract().coresDict)


from tnreason.logic import coordinate_calculus as cc

import numpy as np

alphabet = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
            'u', 'v', 'w', 'x', 'y', 'z']


class BasisCore:
    def __init__(self, core_values, core_colors, headcolor=None, name=None):
        self.values = core_values

        self.colors = core_colors

        if headcolor is None:
            self.headcolor = self.colors[0]
        else:
            self.headcolor = headcolor
        self.name = name

    def clone(self):
        return BasisCore(np.copy(self.values), self.colors.copy(), self.headcolor, self.name)

    def negate(self):
        core0 = self.clone()
        newName = ["not", core0.name]

        newValues = np.tensordot(np.copy(core0.values), create_negation_tensor(),
                                 axes=([core0.colors.index(core0.headcolor)], [0]))

        newColors = core0.colors.copy()
        newColors.pop(newColors.index(core0.headcolor))
        newColors.append(core0.headcolor)

        newHeadColor = core0.headcolor

        return BasisCore(newValues, newColors, newHeadColor, newName)

    def compute_and(self, core1):
        core0 = self.clone()
        newName = [core0.name, "and", core1.name]

        if core0.headcolor == "TruthEvaluated" or core1.headcolor == "TruthEvaluated":
            raise TypeError("Basis Core does not have a Headcolor since already evaluated.")

        contracted = np.tensordot(core0.values, create_and_tensor(), axes=([core0.colors.index(core0.headcolor)], [0]))
        contracted = np.tensordot(contracted, core1.values, axes=([-2], [core1.colors.index(core1.headcolor)]))

        contracted_colors = core0.colors
        contracted_colors.pop(contracted_colors.index(core0.headcolor))
        contracted_colors.append(core0.headcolor)

        core1_colors = core1.colors.copy()
        core1_colors.pop(core1.colors.index(core1.headcolor))

        contracted_colors = contracted_colors + core1_colors

        return BasisCore(contracted, contracted_colors, core0.headcolor, newName)

    def calculate_truth(self):
        truthvec = create_truth_vec()

        if self.headcolor == "TruthEvaluated":
            return ValueError("Already evaluated truth at Basis Core {}!".format(self.name))

        core0 = self.clone()
        contracted = np.tensordot(core0.values, truthvec, axes=([core0.colors.index(core0.headcolor)], [0]))
        contracted_colors = (core0.colors.copy())
        contracted_colors.pop(contracted_colors.index(core0.headcolor))

        return BasisCore(contracted, contracted_colors, "TruthEvaluated", core0.name)

    def reduce_identical_colors(self):
        toBeReduced = []
        for color in self.colors:
            if self.colors.count(color) > 1 and color not in toBeReduced:
                toBeReduced.append(color)
        reducedCore = self.clone()
        for color in toBeReduced:
            reducedCore = reducedCore.reduce_color(color)
        return reducedCore

    def reduce_color(self, reductionColor):
        left_string = ""
        new_string = ""
        for i, color in enumerate(self.colors):
            if color == reductionColor:
                left_string = left_string + alphabet[-1]
            else:
                left_string = left_string + alphabet[i]
                new_string = new_string + alphabet[i]
        new_string = new_string + alphabet[-1]
        right_string = "".join(alphabet[-1] for i in range(self.colors.count(reductionColor)))
        contraction_string = left_string + "," + right_string + "->" + new_string

        delta_values = create_delta_tensor(order=self.colors.count(reductionColor), legDim=2)

        newColors = [color for color in self.colors if color != reductionColor]
        newColors.append(reductionColor)

        return BasisCore(np.einsum(contraction_string, self.values, delta_values), newColors)



    def to_coordinate(self):
        if self.headcolor is not "TruthEvaluated":
            print("Warning: Basis Core to Coordinate Core before Truth Evaluation!")
        return cc.CoordinateCore(self.values, self.colors, self.name)

def create_negation_tensor():
    negation_tensor = np.zeros((2, 2))
    negation_tensor[0, 1] = 1
    negation_tensor[1, 0] = 1
    return negation_tensor


def create_and_tensor():
    and_tensor = np.zeros((2, 2, 2))
    and_tensor[0, 0, 0] = 1
    and_tensor[0, 1, 0] = 1
    and_tensor[1, 0, 0] = 1
    and_tensor[1, 1, 1] = 1
    return and_tensor


# Not required!
def create_delta_tensor(order=3, legDim=2):
    delta_tensor = np.zeros((tuple([2 for k in range(order)])))
    for i in range(legDim):
        delta_tensor[tuple([i for k in range(order)])] = 1
    return delta_tensor


def create_truth_vec():
    truthvec = np.zeros(2)
    truthvec[1] = 1
    return truthvec





from pgmpy.models import MarkovNetwork
from pgmpy.factors.discrete import DiscreteFactor
from pgmpy.inference import VariableElimination
from pgmpy.sampling import GibbsSampling

# from tnreason.logic import expression_calculus as ec
from tnreason.logic import expression_utils as eu

from tnreason.contraction import expression_evaluation as ee

import numpy as np
import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt


class PgmpyMLN:
    def __init__(self, expressionsDict=None):
        self.model = MarkovNetwork()

        if expressionsDict is not None:
            self.extend_from_expressionsDict(expressionsDict)

    def extend_from_expressionsDict(self, expressionsDict):
        factors = []
        edges = []
        for exKey in expressionsDict:
            expression = expressionsDict[exKey][0]
            weight = expressionsDict[exKey][1]
            ex_variables = np.unique(eu.get_variables(expression))

            self.model.add_nodes_from(ex_variables)

            for startnode in ex_variables:
                for endnode in ex_variables:
                    if (startnode, endnode) not in edges and (endnode, startnode) not in edges:
                        if startnode != endnode:
                            edges.append((startnode, endnode))

            core = ee.ExpressionEvaluator(expression, initializeBasisCores=True).create_formula_factor()
            variables = core.colors

            factors.append(DiscreteFactor(variables, [2 for node in variables], np.exp(weight * core.values)))
        self.model.add_edges_from(edges)
        self.model.add_factors(*factors)

    def map_query_given_evidenceDict(self, evidenceDict, variables):
        inference_algorithm = VariableElimination(self.model)
        query_result = inference_algorithm.map_query(evidence=evidenceDict, variables=variables)
        return query_result

    def cond_query_given_evidenceDict(self, evidenceDict, variables):
        inference_algorithm = VariableElimination(self.model)
        query_result = inference_algorithm.query(evidence=evidenceDict, variables=variables)
        return query_result

    def generate_sampleDf(self, sampleNum=1, chainSize=10, method="Gibbs"):
        if method == "Gibbs":
            sampler = GibbsSampling(self.model)
        else:
            return ValueError("Method {} not supported!".format(method))

        df = pd.DataFrame(columns=self.model.nodes)
        for ind in range(sampleNum):
            row_df = pd.DataFrame(sampler.sample(size=chainSize).iloc[-1].to_dict(), index=[ind])
            df = pd.concat([df, row_df])
        return df.astype("int64")

    def visualize(self, regenerate_graph=True, truthDict={}, fontsize=10):
        if regenerate_graph:
            self.graph = nx.Graph()
            self.graph.add_nodes_from(self.model.nodes)
            self.graph.add_edges_from(self.model.edges)
            self.graph_pos = nx.spring_layout(self.graph)
        labels = {atom: atom for atom in self.graph.nodes}
        nx.draw_networkx_labels(self.graph, self.graph_pos, labels, font_size=fontsize)

        colorList = ["r", "g", "b", "r", "g", "b", "r", "g", "b"]
        for i, factor in enumerate(self.model.get_factors()):
            variables = factor.variables
            color = colorList[i]
            edges = [(var1, var2) for var1 in variables for var2 in variables]
            nx.draw_networkx_edges(self.graph, self.graph_pos,
                                   edgelist=edges, width=10, alpha=0.2, edge_color=color)

        trueNodes = [atom for atom in truthDict.keys() if truthDict[atom] == True]
        falseNodes = [atom for atom in truthDict.keys() if truthDict[atom] == False]
        otherNodes = {atom: atom for atom in self.graph.nodes if atom not in truthDict.keys()}

        nx.draw_networkx_nodes(self.graph, self.graph_pos,
                               nodelist=otherNodes,
                               node_color="grey",
                               node_size=3000,
                               alpha=0.2)
        nx.draw_networkx_nodes(self.graph, self.graph_pos,
                               nodelist=trueNodes,
                               node_color="b",
                               node_size=3000,
                               alpha=0.6)
        nx.draw_networkx_nodes(self.graph, self.graph_pos,
                               nodelist=falseNodes,
                               node_color="r",
                               node_size=3000,
                               alpha=0.6)
        plt.show()


if __name__ == "__main__":
    example_expression_dict = {
        "e0": [["not", ["Unterschrank(z)", "and", ["not", "Moebel(z)"]]], 20],
        "e0.5": ["Moebel(z)", 4],
        "e0.625": ["Sledz", 4],
        "e0.626": [["not", ["Moebel(z)", "and", "Sledz"]], 5],
        "e0.75": [["not", [["Unterschrank(z)", "and", "Sledz"], "and", ["not", "Ausgangsrechnung(x)"]]], 2],
        "e1": [["not", "Ausgangsrechnung(x)"], 12],
        "e2": [[["not", "Ausgangsrechnung(x)"], "and", ["not", "Rechnung(x)"]], 14]
    }
    ## 1 = True, 0 = False
    example_evidence_dict = {
        "Unterschrank(z)": 1,
        "Ausgangsrechnung(x)": 1
    }

    test_mln = PgmpyMLN(example_expression_dict)
    cond_query_result = test_mln.cond_query_given_evidenceDict(example_evidence_dict, ["Moebel(z)"])
    cond_query_result.normalize()

    test_mln.visualize(truthDict={"Unterschrank(z)": True, "Ausgangsrechnung(x)": False})



import rdflib
import pandas as pd
import time

def generate_csv_list(ttlPath, limit=None):
    return extend_csv_list(pd.DataFrame(columns=["subject", "predicate", "object"]), ttlPath, limit=limit)


def extend_csv_list(factDf, ttlPath, limit=None):
    g = rdflib.Graph()
    g.parse(ttlPath)

    print("Loading {} successfull.".format(ttlPath))
    print("Graph has {} facts.".format(len(g)))

    reftime = time.time()
    factCount = 0
    for fact in g:
        if time.time() - reftime > 30:
            print("Loaded {} facts.".format(factCount))
            reftime = time.time()
        if limit is not None:
            if factCount > limit:
                return factDf
        rowDf = pd.DataFrame({
            "subject": [str(fact[0])],
            "predicate": [str(fact[1])],
            "object": [str(fact[2])]
            }, index = [factCount])
        factDf = pd.concat([factDf,rowDf])
        factCount += 1
    return factDf


if __name__ == "__main__":

    starttime = time.time()
    df = generate_csv_list("./results/knowledge_graph/parquet_tev2.ttl", limit=50000)
    endtime = time.time()
    print("Generation done after {} seconds.".format(endtime - starttime))

    df.to_csv("./results/csv_files/parquet_tev2_list.csv")
    savetime = time.time()
    print("Saving done after additional {} seconds.".format(savetime - endtime))



import scipy.sparse as scs

class SparseCoordinateCore:

    def __init__(self, core_values, core_colors, name=None):
        self.values = core_values
        self.colors = core_colors
        self.name = name

        if len(core_colors) != len(core_values.shape):
            raise TypeError("Number of Colors provided does not match the Value Shape!")

    ## To be implemented:
    # and, not, sum, solve operations on sparse coordinates


import tnreason.logic.expression_generation as eg

import tnreason.optimization.weight_estimation as wees
import tnreason.optimization.expression_refinement as er

import tnreason.contraction.expression_evaluation as ee

import tnreason.learning.expression_learning as el

from tnreason.model import tensor_network_mln as mln


class SampleBasedMLNLearner:
    def __init__(self, sampleDf=None):
        self.weightedFormulas = []
        if sampleDf is not None:
            self.load_sampleDf(sampleDf)

    def load_sampleDf(self, sampleDf):
        self.sampleDf = sampleDf

    def learn_implication(self, positiveExpression, skeletonExpression, candidatesDict,
                          refinementNum=0, refinementCriterion="weight>1", acceptanceCriterion="weight>0"):
        self.learn(positiveExpression, skeletonExpression, candidatesDict, boostNum=1, saveMod="imp",
                   refinementNum=refinementNum,
                   refinementCriterion=refinementCriterion,
                   acceptanceCriterion=acceptanceCriterion
                   )

    def learn_equivalence(self, positiveExpression, skeletonExpression, candidatesDict,
                          refinementNum=0, refinementCriterion="weight>1", acceptanceCriterion="weight>0"):
        self.learn(positiveExpression, skeletonExpression, candidatesDict, boostNum=1, saveMod="eq",
                   refinementNum=refinementNum,
                   refinementCriterion=refinementCriterion,
                   acceptanceCriterion=acceptanceCriterion
                   )

    def learn_tautology(self, skeletonExpression, candidatesDict,
                        refinementNum=0, refinementCriterion="weight>1", acceptanceCriterion="weight>0"):
        self.learn("Thing", skeletonExpression, candidatesDict, boostNum=1, saveMod="direct",
                   refinementNum=refinementNum,
                   refinementCriterion=refinementCriterion,
                   acceptanceCriterion=acceptanceCriterion
                   )

    def learn(self, positiveExpression, skeletonExpression, candidatesDict,
              boostNum=1, saveMod="eq",
              refinementNum=0, refinementCriterion="weight>1",
              balance=True,
              acceptanceCriterion="weight>0"):
        if saveMod == "eq" or saveMod == "imp":
            # positiveCore = ec.evaluate_expression_on_sampleDf(self.sampleDf, positiveExpression)
            positiveCore = ee.ExpressionEvaluator(positiveExpression, sampleDf=self.sampleDf).evaluate()
        else:
            # positiveCore = ec.evaluate_expression_on_sampleDf(self.sampleDf, "Thing")
            positiveCore = ee.ExpressionEvaluator("Thing", sampleDf=self.sampleDf).evaluate()

        solutionExpressions = self.boost(skeletonExpression, candidatesDict, positiveCore, boostNum=boostNum,
                                         refinementNum=refinementNum, refinementCriterion=refinementCriterion,
                                         balance=balance)
        ## Adding the formulas to the model
        for learnedPremise in solutionExpressions:
            if saveMod == "eq" or saveMod == "imp":
                solutionExpression = eg.generate_from_generic_expression([learnedPremise, saveMod, positiveExpression])
                self.add_independent_formula(solutionExpression, criterion=acceptanceCriterion)
            else:
                self.add_independent_formula(learnedPremise, criterion=acceptanceCriterion)

    def boost(self, skeletonExpression, candidatesDict, positiveCore, boostNum=1, refinementNum=0,
              refinementCriterion="weight>1",
              balance=True):
        solutionExpressions = []
        for boostPos in range(boostNum):
            print("## Boost number {} ##".format(boostPos))
            refiningExpression = "Thing"
            for solutionExpression in solutionExpressions:
                refiningExpression = [refiningExpression, "and", ["not", solutionExpression]]
            # refiningCore = ec.evaluate_expression_on_sampleDf(self.sampleDf, refiningExpression)
            refiningCore = ee.ExpressionEvaluator(refiningExpression, sampleDf=self.sampleDf).evaluate()
            positiveCore = positiveCore.compute_and(refiningCore)
            negativeCore = positiveCore.negate()

            solutionExpressions.append(self.refine(skeletonExpression, candidatesDict, positiveCore, negativeCore,
                                                   refinementLeft=refinementNum,
                                                   acceptanceCriterion=refinementCriterion,
                                                   balance=balance))
        return solutionExpressions

    def refine(self, skeletonExpression, candidatesDict, positiveCore, negativeCore,
               refinementLeft=0, acceptanceCriterion="weight>0.5",
               balance=True):
        solutionExpression = self.optimize_formula(skeletonExpression, candidatesDict, positiveCore, negativeCore,
                                                   balance=balance)
        empRate, satRate, weight = self.learn_independent_weight(solutionExpression)
        if refinementLeft > 0 and not criterion_satisfied(empRate, satRate, weight, acceptanceCriterion):
            print("# Refining since criterion not satisfied, {} tries left #".format(refinementLeft))
            newSkeleton = er.add_leaf_atom(skeletonExpression)
            refinementLeft -= 1
            return self.refine(newSkeleton, candidatesDict, positiveCore, negativeCore,
                               refinementLeft)
        else:
            print("# Solution is {} #".format(solutionExpression))
            return solutionExpression

    def optimize_formula(self, skeletonExpression, candidatesDict, positiveCore, negativeCore,
                         optimizationInstructions=["als2", "project", "als2", "project", "als2", "project"],
                         balance=True):
        exLearner = el.SampleBasedOptimizer(skeletonExpression, candidatesDict)

        exLearner.generate_fixedCores_sampleDf(self.sampleDf)
        exLearner.generate_target_and_filterCore_from_exampleCores(positiveCore, negativeCore)
        if balance:
            exLearner.balance_importance(positiveCore=positiveCore, negativeCore=negativeCore)
        exLearner.random_initialize_variableCoresDict()
        for optInstruction in optimizationInstructions:
            if optInstruction.startswith("als"):
                exLearner.als(int(optInstruction[3:]))
            if optInstruction == "project":
                exLearner.get_solution(adjustVariablesCoresDict=True)
        exLearner.get_solution()

        return exLearner.solutionExpression

    def add_independent_formula(self, expression, criterion=""):
        empRate, satRate, weight = self.learn_independent_weight(expression)
        if criterion_satisfied(empRate, satRate, weight, criterion):
            self.weightedFormulas.append([expression, weight])
            print("The expression {} with weight {} passed criterion {} and has been added.".format(expression, weight,
                                                                                                    criterion))
        else:
            print(
                "Expression {} with weight {} does not satisfy the criterion {}.".format(expression, weight, criterion))

    def learn_independent_weight(self, expression, verbose=False):
        return wees.calculate_weight(expression, self.sampleDf, verbose=verbose)

    def alternating_weight_optimization(self, sweepNum):
        estimator = wees.WeightEstimator([weightedFormula[0] for weightedFormula in self.weightedFormulas],
                                         sampleDf=self.sampleDf)
        estimator.alternating_optimization(sweepNum)
        self.weightedFormulas = [[estimator.formulaDict[formulaKey][0], estimator.formulaDict[formulaKey][3]] for
                                 formulaKey in estimator.formulaDict]

    def generate_mln(self):
        return mln.TensorMLN(expressionsDict=
                             {str(i): [self.weightedFormulas[i][0], self.weightedFormulas[i][1]]
                              for i in range(len(self.weightedFormulas))})


## Older name of the class
AtomicMLNLearner = SampleBasedMLNLearner


def criterion_satisfied(empRate, satRate, weight, criterion):
    criteria = criterion.split(",")
    for criterion in criteria:
        if criterion.startswith("weight"):
            threshold = float(criterion.split(">")[1])
            if weight < threshold:
                return False
        elif criterion.startswith("empRate"):
            threshold = float(criterion.split(">")[1])
            if empRate < threshold:
                return False
        elif criterion.startswith("satRate"):
            threshold = float(criterion.split("<")[1])
            if satRate > threshold:
                return False
        else:
            raise ValueError("Acceptance Criterion {} not understood.".format(criterion))
    return True



import pandas as pd

def sampleDf_to_pairDf(sampleDf,prefix):
    columns = sampleDf.columns
    dataNum = len(sampleDf)
    rootClasses = find_rootClasses(columns)
    pairDf = pd.DataFrame(data= {entry:[prefix+str(entry)+"_"+str(i) for i in range(dataNum)] for entry in rootClasses},
                          columns=rootClasses)
    return pairDf

def find_rootClasses(columns):
    rootClasses = []
    for column in columns:
        argpart = column.split("(")[1][:-1]
        if "," in column:
            arg1, arg2 = argpart.split(",")
            if arg1 not in rootClasses:
                rootClasses.append(arg1)
            if arg2 not in rootClasses:
                rootClasses.append(arg2)
        else:
            if argpart not in rootClasses:
                rootClasses.append(argpart)
    return rootClasses


import numpy as np

def get_all_variables(expressionList):
    variables = []
    for expression in expressionList:
        variables = variables + get_variables(expression)
    return np.unique(variables)


def get_variables(expression):
    if type(expression) == str:
        return [expression]
    elif expression[0] == "not":
        return get_variables(expression[1])
    elif expression[1] == "and":
        if type(expression[0]) == str:
            left_variables = [expression[0]]
        else:
            left_variables = get_variables(expression[0])
        if type(expression[2]) == str:
            right_variables = [expression[2]]
        else:
            right_variables = get_variables(expression[2])
        return left_variables + right_variables
    else:
        raise ValueError("Expression {} not understood.".format(expression))

def get_individuals(expression):
    if type(expression) == str:
        arguments = expression.split("(")[1][:-1]
        if "," in arguments:
            return arguments.split(",")
        else:
            return [arguments]
    elif expression[0] == "not":
        return get_individuals(expression[1])
    elif expression[1] == "and":
        return list(np.unique(np.concatenate((get_individuals(expression[0]), get_individuals(expression[2])))))


from tnreason.logic import coordinate_calculus as cc

import numpy as np


## When only atoms in expressions (FormulaTensor)
def create_subExpressionCores(expression, formulaKey):
    addCoreKey = str(formulaKey) + "_" + str(expression) + "_subCore"
    if type(expression) == str:
        return {addCoreKey: cc.CoordinateCore(np.eye(2), [expression, formulaKey + "_" + expression], expression)}
    elif expression[0] == "not":
        if type(expression[1]) == str:
            return {addCoreKey: cc.CoordinateCore(create_negation_tensor(),
                                                  [expression[1], formulaKey + "_" + str(expression)],
                                                  expression)}
        else:
            partsDict = create_subExpressionCores(expression[1], formulaKey)
            partsDict[addCoreKey] = cc.CoordinateCore(create_negation_tensor(),
                                                      [formulaKey + "_" + str(expression[1]),
                                                       formulaKey + "_" + str(expression)], expression)
            return partsDict
    elif expression[1] == "and":
        if type(expression[0]) == str:
            partsDict0 = {}
            leftColor = expression[0]
        else:
            partsDict0 = create_subExpressionCores(expression[0], formulaKey)
            leftColor = formulaKey + "_" + str(expression[0])

        if type(expression[2]) == str:
            partsDict2 = {}
            rightColor = expression[2]
        else:
            partsDict2 = create_subExpressionCores(expression[2], formulaKey)
            rightColor = formulaKey + "_" + str(expression[2])

        return {**partsDict0, **partsDict2,
                addCoreKey: cc.CoordinateCore(create_and_tensor(),
                                              [leftColor, rightColor, formulaKey + "_" + str(expression)],
                                              str(expression))}


def create_truth_vec():
    truthvec = np.zeros(2)
    truthvec[1] = 1
    return truthvec


def create_negation_tensor():
    negation_tensor = np.zeros((2, 2))
    negation_tensor[0, 1] = 1
    negation_tensor[1, 0] = 1
    return negation_tensor


def create_and_tensor():
    and_tensor = np.zeros((2, 2, 2))
    and_tensor[0, 0, 0] = 1
    and_tensor[0, 1, 0] = 1
    and_tensor[1, 0, 0] = 1
    and_tensor[1, 1, 1] = 1
    return and_tensor


def create_headCore(headType, weight, headColor):
    if headType == "truthEvaluation":
        headValues = np.zeros(shape=(2))
        headValues[1] = 1  # weight
    elif headType == "weightedTruthEvaluation":
        headValues = np.zeros(shape=(2))
        headValues[1] = weight
    elif headType == "expFactor":
        headValues = create_expFactor_values(weight, False)
    elif headType == "diffExpFactor":
        headValues = create_expFactor_values(weight, True)
    else:
        raise ValueError("Headtype {} not understood!".format(headType))
    return cc.CoordinateCore(headValues, [headColor])


def create_expFactor_values(weight, differentiated):
    values = np.zeros(shape=(2))
    if not differentiated:
        values[0] = 1
    values[1] = np.exp(weight)
    return values


def create_evidenceCoresDict(evidenceDict):
    evidenceCoresDict = {}
    for atomKey in evidenceDict:
        truthValues = np.zeros(shape=(2))
        if bool(evidenceDict[atomKey]):
            truthValues[1] = 1
        else:
            truthValues[0] = 1
        evidenceCoresDict[atomKey + "_evidence"] = cc.CoordinateCore(truthValues, [atomKey], atomKey + "_evidence")
    return evidenceCoresDict


## When Placeholders in Expression (SuperposedFormulaTensor)
def skeleton_recursion(headExpression, candidatesDict):
    if type(headExpression) == str:
        return {}, candidatesDict[headExpression]
    elif headExpression[0] == "not":
        if type(headExpression[1]) == str:
            return create_negationCoreDict(candidatesDict[headExpression[1]], inprefix=str(headExpression[1]) + "_",
                                           outprefix=str(headExpression) + "_"), candidatesDict[headExpression[1]]
        else:
            skeletonCoresDict, atoms = skeleton_recursion(headExpression[1], candidatesDict)
            return {**skeletonCoresDict,
                    **create_negationCoreDict(atoms, inprefix=str(headExpression[1]) + "_",
                                              outprefix=str(headExpression)) + "_"}, atoms
    elif headExpression[1] == "and":
        if type(headExpression[0]) == str:
            leftskeletonCoresDict = {headExpression[0] + "_" + atomKey + "_l": create_deltaCore(
                colors=[headExpression[0] + "_" + atomKey, str(headExpression) + "_" + atomKey])
                for atomKey in candidatesDict[headExpression[0]]}
            leftatoms = candidatesDict[headExpression[0]]
        else:
            leftskeletonCoresDict, leftatoms = skeleton_recursion(headExpression[0], candidatesDict)

            leftskeletonCoresDict = {**leftskeletonCoresDict,
                                     **{str(headExpression[0]) + "_" + atomKey + "_lPass": create_deltaCore(
                                         [str(headExpression[0]) + "_" + atomKey, str(headExpression) + "_" + atomKey])
                                         for atomKey in leftatoms}
                                     }
        if type(headExpression[2]) == str:
            rightskeletonCoresDict = {headExpression[2] + "_" + atomKey + "_r": create_deltaCore(
                colors=[headExpression[2] + "_" + atomKey, str(headExpression) + "_" + atomKey])
                for atomKey in candidatesDict[headExpression[2]]}
            rightatoms = candidatesDict[headExpression[2]]
        else:
            rightskeletonCoresDict, rightatoms = skeleton_recursion(headExpression[2], candidatesDict)
            rightskeletonCoresDict = {**rightskeletonCoresDict,
                                      **{str(headExpression[2]) + "_" + atomKey + "_rPass": create_deltaCore(
                                          [str(headExpression[2]) + "_" + atomKey, str(headExpression) + "_" + atomKey])
                                          for atomKey in rightatoms}
                                      }
        return {**leftskeletonCoresDict, **rightskeletonCoresDict}, leftatoms + rightatoms


def create_negationCoreDict(atoms, inprefix, outprefix):
    negationMatrix = np.zeros(shape=(2, 2))
    negationMatrix[0, 1] = 1
    negationMatrix[1, 0] = 1

    negationCoreDict = {}
    for atomKey in atoms:
        negationCoreDict[outprefix + "_" + atomKey + "_neg"] = cc.CoordinateCore(negationMatrix,
                                                                                 [inprefix + atomKey,
                                                                                  outprefix + atomKey],
                                                                                 outprefix + atomKey + "_neg")
    return negationCoreDict


def create_deltaCore(colors, name=""):
    values = np.zeros(shape=[2 for i in range(len(colors))])
    values[tuple(0 for color in colors)] = 1
    values[tuple(1 for color in colors)] = 1
    return cc.CoordinateCore(values, colors, name)


def create_selectorCoresDict(candidatesDict):
    ## incolors: placeHolderKey
    ## outcolors: placeHolderKey + "_" + atomKey
    selectorCoresDict = {}
    for placeHolderKey in candidatesDict:
        for i, atomKey in enumerate(candidatesDict[placeHolderKey]):
            coreValues = np.ones(shape=(len(candidatesDict[placeHolderKey]), 2))
            coreValues[i, 0] = 0
            selectorCoresDict[placeHolderKey + "_" + atomKey + "_selector"] = cc.CoordinateCore(
                coreValues, [placeHolderKey, placeHolderKey + "_" + atomKey],
                placeHolderKey + "_" + atomKey + "_selector")
    return selectorCoresDict


## DataCore Creation
def dataCore_from_sampleDf(sampleDf, atomKey, dataColor):
    if atomKey not in sampleDf.keys():
        raise ValueError
    dfEntries = sampleDf[atomKey].values
    dataNum = dfEntries.shape[0]
    values = np.zeros(shape=(dataNum, 2))
    for i in range(dataNum):
        if dfEntries[i] == 0:
            values[i, 0] = 1
        else:
            values[i, 1] = 1
    return cc.CoordinateCore(values, [dataColor, atomKey])



## Analysis

## Check whether the colors in all coreDicts match wrt each other and the knownShapesDict
def check_colorShapes(coresDicts, knownShapesDict={}):
    for coresDict in coresDicts:
        for coreKey in coresDict:
            for i, color in enumerate(coresDict[coreKey].colors):
                coreColorShape = coresDict[coreKey].values.shape[i]
                if color not in knownShapesDict:
                    knownShapesDict[color] = coreColorShape
                else:
                    if knownShapesDict[color] != coreColorShape:
                        raise ValueError("Core {} has unexpected shape of color {}.".format(coreKey, color))



import numpy as np

from tnreason.contraction import core_contractor as coc
from tnreason.model import tensor_model as tm
from tnreason.model import formula_tensors as ft


def expected_cross_entropy(testExpressionsDict, generativeExpressionsDict):
    testTensorModel = tm.TensorRepresentation(testExpressionsDict, headType="expFactor")
    expGenerativeTensorModel = tm.TensorRepresentation(generativeExpressionsDict, headType="expFactor")

    testPartition = testTensorModel.contract_partition()
    generativePartition = expGenerativeTensorModel.contract_partition()

    crossTerm = np.sum(
        coc.CoreContractor({**testTensorModel.get_cores([formulaKey], headType="weightedTruthEvaluation"),
            **expGenerativeTensorModel.all_cores()}).contract().values
        for formulaKey in testExpressionsDict
    )

    return np.log(testPartition) - crossTerm / generativePartition


def expected_shannon_entropy(testExpressionsDict):
    return expected_cross_entropy(testExpressionsDict, testExpressionsDict)


def expected_KL_divergence(testExpressionsDict, generativeExpressionsDict):
    return expected_cross_entropy(testExpressionsDict, generativeExpressionsDict) - expected_shannon_entropy(
        generativeExpressionsDict)


## Further entropies:
# empirical_cross_entropy: computed by the likelihood in MLE Base
# empirical_shannon_entropy: in ft.DataTensor
# empirical_KL_divergence: difference of likelihood with empirical shannon entropy (also done in MLE Base)

def empirical_shannon_entropy(sampleDf, atoms=None):
    return ft.DataTensor(sampleDf,atoms).compute_shannon_entropy()


from tnreason.logic import expression_utils as eu

from tnreason.contraction import core_contractor as coc
from tnreason.model import create_cores as crc

import numpy as np


class FormulaTensor:
    '''
    expression:
    formulaKey: For distinction , typically passed by TensorRepresentation as dict key
    headType: Whether \weight\ftensor (truthEvaluation, default) or \expof{\weight\ftensor} (expFactor) or \partial\expof{\weight\ftensor} (diffExpFactor) is created
    weight: Factor on the formulaTensor
    '''

    def __init__(self, expression, formulaKey=None, headType="weightedTruthEvaluation", weight=1):
        if formulaKey is not None:
            self.formulaKey = formulaKey
        else:
            self.formulaKey = str(expression)
        self.expression = expression
        self.atoms = eu.get_variables(expression)

        ## Build the Cores
        self.create_subExpressionCores()
        self.set_head(headType=headType, weight=weight)
        self.weight = weight

    def create_subExpressionCores(self):
        self.subExpressionCoresDict = crc.create_subExpressionCores(self.expression, str(self.formulaKey))

    def set_head(self, headType, weight=1):
        self.weight = weight
        self.headCore = crc.create_headCore(headType, weight, headColor=self.formulaKey + "_" + str(self.expression))

    def infer_on_evidenceDict(self, evidenceDict):
        return coc.CoreContractor({**self.get_cores(), **crc.create_evidenceCoresDict(evidenceDict)},
                                  openColors=[atomKey for atomKey in self.atoms if
                                              atomKey not in evidenceDict]).contract()

    def get_cores(self):
        return {**self.subExpressionCoresDict, self.formulaKey + "_head": self.headCore}


class SuperposedFormulaTensor:
    ## Shall be the central object to be optimized during MLE
    # Gradient is just the omission of the respective parameterCore

    def __init__(self, skeletonExpression, candidatesDict, parameterCoresDict=None):
        self.skeletonExpression = skeletonExpression
        self.candidatesDict = candidatesDict

        if parameterCoresDict is not None:  # former variableCoresDict
            self.set_parameterCoresDict(parameterCoresDict)
        self.create_selectorCoresDict()
        self.create_skeletonCoresDict()

        self.dataCoresDict = {}

    def set_parameterCoresDict(self, parameterCoresDict):
        self.parameterCoresDict = parameterCoresDict
        ## Check for consistency with the candidatesDict
        crc.check_colorShapes([parameterCoresDict],
                              knownShapesDict={key: len(self.candidatesDict[key]) for key in self.candidatesDict})

    def random_initialize_parameterCoresDict(self):
        for coreKey in self.parameterCoresDict:
            self.parameterCoresDict[coreKey].values = np.random.random(
                size=self.parameterCoresDict[coreKey].values.shape)

    def get_largest_weight_as_solutionMap(self):
        ## Choose formula Tensor by maximal entry
        contractedParameters = coc.CoreContractor(self.parameterCoresDict,
                                                  openColors=self.candidatesDict.keys()).contract()
        maxPos = np.argmax(contractedParameters.values)
        maxIndices = np.unravel_index(maxPos, contractedParameters.values.shape)
        solutionMap = {
            str(contractedParameters.colors[i]): self.candidatesDict[str(contractedParameters.colors[i])][maxIndices[i]]
            for i in range(len(maxIndices))}
        return solutionMap

    def create_selectorCoresDict(self):
        ## incolors: placeHolderKey
        ## outcolors: placeHolderKey + "_" + atomKey
        self.selectorCoresDict = crc.create_selectorCoresDict(self.candidatesDict)

    def create_skeletonCoresDict(self):
        ## incolors: placeHolderKey + "_" + atomKey
        ## outcolors: atomKey
        self.skeletonCoresDict, self.atoms = crc.skeleton_recursion(self.skeletonExpression, self.candidatesDict)
        for atomKey in self.atoms:
            self.skeletonCoresDict[atomKey + "_skeletonHeadCore"] = crc.create_deltaCore(
                [str(self.skeletonExpression) + "_" + atomKey, atomKey], atomKey + "_skeletonHeadCore")

    def get_cores(self, parameterExceptionKeys=[]):
        return {**{key: self.parameterCoresDict[key] for key in self.parameterCoresDict if
                   key not in parameterExceptionKeys},
                **self.selectorCoresDict,
                **self.skeletonCoresDict}


class DataTensor:
    def __init__(self, sampleDf, atoms=None, dataColor="j"):
        if atoms is None:
            self.atoms = sampleDf.columns
        else:
            self.atoms = atoms
        self.dataColor = dataColor

        self.create_dataCores(sampleDf)

    def create_dataCores(self, sampleDf):
        self.dataNum = sampleDf.values.shape[0]
        self.dataCores = {
            atomKey + "_data": crc.dataCore_from_sampleDf(sampleDf, atomKey, self.dataColor)
            for atomKey in self.atoms
        }

    def compute_shannon_entropy(self):
        contractedData = coc.CoreContractor(self.dataCores,
                                            openColors=self.atoms).contract().values.flatten() / self.dataNum
        logContractedData = np.log(np.copy(contractedData))
        logContractedData[logContractedData < -1e308] = 0
        return -np.dot(logContractedData, contractedData)

    def get_cores(self):
        return self.dataCores


if __name__ == "__main__":
    from tnreason.contraction import contraction_visualization as cv

    expression = ["A1", "and", ["not", "A2"]]
    fTensor = FormulaTensor(expression, "f1", headType="truthEvaluation", weight=1)
    print(fTensor.infer_on_evidenceDict({"A1": 1}).values)
    cv.draw_contractionDiagram({**fTensor.subExpressionCoresDict, "head": fTensor.headCore})



def generate_list_from_rule(premises, head):
    expression = premises[0]
    for premise in premises[1:]:
        expression = [premise, "and", expression]
    expression = ["not", [["not", head], "and", expression]]

    return expression


def generate_conjunctions(atoms):
    expression = atoms[0]
    for atom in atoms[1:]:
        expression = [atom, "and", expression]
    return expression


def generate_negated_conjunctions(positive_atoms, negated_atoms):
    return generate_conjunctions(positive_atoms + [["not", atom] for atom in negated_atoms])


def generate_from_generic_expression(expression):
    if type(expression) == str:
        return expression
    elif expression[0] == "not":
        return ["not", generate_from_generic_expression(expression[1])]
    elif expression[1] == "and":
        return [generate_from_generic_expression(expression[0]), "and", generate_from_generic_expression(expression[2])]
    elif expression[1] == "or":
        return ["not", [["not", generate_from_generic_expression(expression[0])], "and",
                        ["not", generate_from_generic_expression(expression[2])]]]
    elif expression[1] == "eq":
        left = generate_from_generic_expression(expression[0])
        right = generate_from_generic_expression(expression[2])
        return ["not", [
            ["not", [left, "and", right]]
            , "and",
            ["not", [["not", left], "and", ["not", right]]]
        ]]
    elif expression[1] == "imp":
        left = generate_from_generic_expression(expression[0])
        right = generate_from_generic_expression(expression[2])
        return ["not", [
            left, "and", ["not", right]
        ]]
    else:
        raise ValueError("Expression {} not understood!".format(expression))


def replace_atoms(expression, atomDict):
    if type(expression) == str:
        return atomDict[expression]
    else:
        if expression[0] == "not":
            return ["not", replace_atoms(expression[1], atomDict)]
        elif expression[1] == "and":
            return [replace_atoms(expression[0], atomDict), "and", replace_atoms(expression[2], atomDict)]


def generate_pracmln_string(expression, weight):
    return str(weight) + " " + generate_pracmln_formulastring(expression)


def generate_pracmln_formulastring(expression):
    if type(expression) == str:
        return expression
    elif expression[0] == "not":
        return "!(" + generate_pracmln_string(expression[1]) + ")"
        # raise TypeError("pracmln string model does not yet support {}.".format(expression))
    elif expression[1] == "and":
        return generate_pracmln_string(expression[0]) + " ^ " + generate_pracmln_string(expression[2])


if __name__ == "__main__":
    and_expression = generate_from_generic_expression(["jaszczur", "or", ["sikorka", "or", ["not", "sledz"]]])
    print(and_expression)



import rdflib
import pandas as pd
import time

def generate_ttl(csvPath, limit=None):
    dataFrame = pd.read_csv(csvPath)
    g = dataframe_to_kg(dataFrame)
    return g

def dataframe_to_kg(dataFrame):
    g = rdflib.Graph()
    for i, row in dataFrame.iterrows():
        subject = rdflib.URIRef(row["subject"])
        predicate = rdflib.URIRef(row["predicate"])
        object = rdflib.URIRef(row["object"])
        g.add((subject,predicate,object))
    return g


import numpy as np
import matplotlib.pyplot as plt

from tnreason.logic import expression_generation as eg, optimization_calculus as oc


class GeneralizedALS:
    def __init__(self, variableCoresDict, fixedCoresDict):
        self.variableCoresDict = variableCoresDict  # Dictionaries (!) of Coordinate Cores ## Needed to be initialized to know what legs are contracted
        self.fixedCoresDict = fixedCoresDict
        self.filterCore = None

    def set_targetCore(self, targetCore):
        self.targetCore = targetCore

    def set_filterCore(self, filterCore):
        self.filterCore = filterCore

    def sweep(self, sweepnum=1, contractionScheme=None, verbose=False):
        residua = np.empty((sweepnum, len(self.variableCoresDict)))

        if np.linalg.norm(self.targetCore.values)==0 and verbose:
            print("Warning: Target Core is Zero!")

        if contractionScheme is None:
            contractionScheme = eg.generate_conjunctions([key for key in self.variableCoresDict.keys()])

        #if verbose:
        #    for key in self.variableCoresDict:
        #        print("Before Sweeping")
        #        print(key, self.variableCoresDict[key].values)
        #res_remember = None
        for i in range(sweepnum):
            if verbose:
                print("## SWEEP {} ##".format(i))
            for k, legKey in enumerate(self.variableCoresDict):
                residua[i, k] = self.leg_optimization(legKey, contractionScheme=contractionScheme,verbose=verbose)
                if verbose:
                    print("Optimized leg {}: Residuum is {}".format(legKey, residua[i, k]))

                #assert (res_remember is None or residua[i, k]-res_remember < 0.0001 * np.linalg.norm(self.targetCore.values)), (
                #                "Residuum increased on leg {}!".format(legKey))
                #res_remember = residua[i, k]

            ## STABILIZE LEGS MAKES PROBLEMS WHEN HAVING NOTS (AFFINE LINEARITIES)!
            # self.stabilize_legs()

        if verbose:
            print("After Sweeping")
            for key in self.variableCoresDict:
                print(key,self.variableCoresDict[key].values)
        self.residua = residua

    def stabilize_legs(self):
        normproduct = 1
        for legKey in self.variableCoresDict:
            normproduct = normproduct * np.linalg.norm(self.variableCoresDict[legKey].values)
        normsoll = normproduct ** (1 / len(self.variableCoresDict))
        for legKey in self.variableCoresDict:
            self.variableCoresDict[legKey].values = self.variableCoresDict[legKey].values \
                                                    * (normsoll / np.linalg.norm(self.variableCoresDict[legKey].values))

    ## Limited to situation, where each datacore has a legcore with same key
    # contractionScheme is a nested list determining contraction to be used in a customized way, default: conjunctions
    def leg_optimization(self, legKey, contractionScheme, verbose):
        core_dict = oc.calculate_core_dict(self.variableCoresDict, self.fixedCoresDict, legKey)

        #operator, constant = ec.calculate_operator_and_constant(core_dict, contractionScheme, legKey,
        #                                                        self.variableCoresDict[legKey].colors)
        coreList = oc.calculate_coreList(core_dict, contractionScheme, self.variableCoresDict[legKey].colors)
        operator, constant = oc.operator_constant_from_coreList(coreList, self.variableCoresDict[legKey].colors)

        if self.filterCore is not None:
            operator = operator.compute_and(self.filterCore)
            constant = constant.compute_and(self.filterCore)
            if np.linalg.norm(operator.values) == 0 and verbose:
                print("Warning: Operator is Zero! Possible reasons: Zero targetCore resulted in zero solution, "
                      "filterCore is orthogonal to prefiltered operator.")

        constant = constant.negate(ignore_ones=True)
        correctedTarget = self.targetCore.sum_with(constant).compute_and(self.filterCore)

        sol = solve_lstsq(operator, correctedTarget, self.variableCoresDict[legKey].colors)
        self.variableCoresDict[legKey].values = sol

        residuum = calculate_residuum(operator, self.variableCoresDict[legKey].clone(), correctedTarget)
        return np.linalg.norm(residuum.values)

    def visualize_residua(self):
        plt.imshow(self.residua, cmap="coolwarm", vmin=0, vmax=np.linalg.norm(self.targetCore.values))

        plt.colorbar()
        plt.title("Residuum Norm Decay", fontsize=20)
        plt.xlabel("Legs")
        plt.ylabel("Iterations")

        plt.show()

def solve_lstsq(X_k_core, y_core, leg_colors):
    lhs_core = X_k_core.contract_common_colors(X_k_core, exceptions =leg_colors)
    lhs_core.reorder_colors(leg_colors + [color + "_cor" for color in leg_colors])
    variables_leg_shapes = lhs_core.values.shape[:int(len(lhs_core.values.shape)/2)]
    var_product = np.prod(variables_leg_shapes)

    lhs = lhs_core.values

    rhs_core = X_k_core.clone().contract_common_colors(y_core.clone())
    rhs = rhs_core.values

    lhs_reshaped = lhs.reshape(var_product,var_product)
    rhs_reshaped = rhs.reshape(var_product)

    solution, res, rank, s = np.linalg.lstsq(lhs_reshaped, rhs_reshaped, rcond=None)
    return solution.reshape(variables_leg_shapes)


def calculate_residuum(operator_core, solution_core, y_core):
    prediction = operator_core.clone().contract_common_colors(solution_core.clone())
    prediction = prediction.negate(ignore_ones=True)
    return prediction.sum_with(y_core.clone())


def create_core_dict(variableCoresDict, fixedCoresDict, legKey=None):
    core_dict = {}
    for coreKey in fixedCoresDict:
        if coreKey != legKey:
            core_dict[coreKey] = fixedCoresDict[coreKey].clone().contract_common_colors(
                variableCoresDict[coreKey].clone())
        else:
            core_dict[coreKey] = fixedCoresDict[coreKey].clone()
    return core_dict



import numpy as np

from tnreason.logic import expression_utils as eu
from tnreason.logic import coordinate_calculus as cc
from tnreason.logic import basis_calculus as bc


class ExpressionEvaluator:
    def __init__(self, expression, atomDict=None, sampleDf=None, initializeBasisCores=False):
        self.expression = expression
        if sampleDf is not None:
            self.create_atomDict_from_sampleDf(sampleDf)
        if atomDict is not None:
            self.atomDict = atomDict
        if initializeBasisCores:
            self.create_atomDict_basisCores()

    def create_atomDict_from_sampleDf(self, sampleDf):
        atomKeys = np.unique(eu.get_variables(self.expression))
        self.atomDict = {}
        for atomKey in atomKeys:
            if atomKey == "Thing":
                values = np.ones(sampleDf.shape[0])
            elif atomKey == "Nothing":
                values = np.zeros(sampleDf.shape[0])
            else:
                values = sampleDf[atomKey].astype("int64").values
            self.atomDict[atomKey] = cc.CoordinateCore(values, ["j"], atomKey)

    def create_atomDict_basisCores(self):
        atomKeys = np.unique(eu.get_variables(self.expression))
        self.atomDict = {}
        for atomKey in atomKeys:
            ## How to include Thing/Nothing atomKeys?
            self.atomDict[atomKey] = bc.BasisCore(np.eye(2), [atomKey, "head"], headcolor="head", name=atomKey)

    def evaluate(self):
        return self.calculate_core(self.expression)

    def evaluate_on_sampleDf(self, sampleDf):
        self.create_atomDict_from_sampleDf(sampleDf)
        return self.calculate_core(self.expression)

    def create_formula_factor(self):
        return self.calculate_core(self.expression).calculate_truth().reduce_identical_colors().to_coordinate()

    ## Included here from expression_calculus: Benefit is avoiding copies of the atomDict.
    def calculate_core(self, expression):
        if type(expression) == str:
            return self.atomDict[expression]
        elif expression[0] == "not":
            return self.calculate_core(expression[1]).negate()
        elif expression[1] == "and":
            return self.calculate_core(expression[0]).compute_and(
                self.calculate_core(expression[2]))
        else:
            raise ValueError("Expression {} not understood.".format(expression))


if __name__ == "__main__":
    coordinateDict = {
        "a": cc.CoordinateCore(np.random.binomial(n=1, p=0.4, size=(10, 7, 5)), ["x", "y", "z"], name="a"),
        "b": cc.CoordinateCore(np.random.binomial(n=1, p=0.4, size=(10, 7, 5)), ["x", "q", "z"], name="b"),
    }

    # expression = ["a", "and", "b"]
    expression = ["not", "a"]

    evaluator = ExpressionEvaluator(expression, atomDict=coordinateDict)
    result = evaluator.evaluate()
    print(result.colors, result.values.shape, np.sum(result.values))



from tnreason.model import formula_tensors as ft
from tnreason.model import model_visualization as mv

from tnreason.logic import expression_utils as eu
from tnreason.logic import coordinate_calculus as cc

from tnreason.contraction import core_contractor as coc

import numpy as np


class TensorRepresentation:
    def __init__(self, expressionsDict, headType="expFactor"):
        self.formulaTensorsDict = {formulaKey: ft.FormulaTensor(expression=expressionsDict[formulaKey][0],
                                                                formulaKey=formulaKey,
                                                                headType=headType,
                                                                weight=expressionsDict[formulaKey][1]
                                                                ) for formulaKey in expressionsDict}
        self.expressionsDict = expressionsDict
        self.headType = headType  ## To prevent resetting of headCores

        for formulaKey in expressionsDict:
            self.formulaTensorsDict[formulaKey].set_head(headType, weight=expressionsDict[formulaKey][1])

        self.atoms = np.unique(eu.get_all_variables([expressionsDict[formulaKey][0] for formulaKey in expressionsDict]))

    def add_expression(self, expression, weight=1, formulaKey=None):
        if formulaKey is None:
            formulaKey = str(expression)
        self.formulaTensorsDict[formulaKey] = ft.FormulaTensor(expression=expression, formulaKey=formulaKey,
                                                               headType=self.headType,
                                                               weight=weight)
        self.expressionsDict[formulaKey] = [expression, weight]
        self.atoms = np.unique(self.atoms, eu.get_variables(expression))

    def all_cores(self):
        allCoresDict = {}
        for formulaKey in self.formulaTensorsDict:
            allCoresDict = {**allCoresDict, **self.formulaTensorsDict[formulaKey].subExpressionCoresDict,
                            formulaKey + "_head": self.formulaTensorsDict[formulaKey].headCore}
        return allCoresDict

    def get_cores(self, formulaKeys=None, headType="expFactor"):
        if formulaKeys is None:
            formulaKeys = self.formulaTensorsDict.keys()
        if self.headType != headType:
            self.set_heads(headType)
        retCoresDict = {}
        for formulaKey in formulaKeys:
            retCoresDict = {**retCoresDict, **self.formulaTensorsDict[formulaKey].get_cores()}
        return retCoresDict

    def set_heads(self, headType):
        if self.headType != headType:
            print("Setting to {}".format(headType))
            for formulaKey in self.formulaTensorsDict.keys():
                self.formulaTensorsDict[formulaKey].set_head(headType, weight=self.expressionsDict[formulaKey][1])
            self.headType = headType

    def update_heads(self, updateDict, headType=None):
        if headType is None:
            headType = self.headType
        for formulaKey in updateDict:
            self.expressionsDict[formulaKey][1] = updateDict[formulaKey]
            self.formulaTensorsDict[formulaKey].set_head(headType, weight=updateDict[formulaKey])

    def get_weights(self):
        return {formulaKey: self.formulaTensorsDict[formulaKey].weight for formulaKey in self.formulaTensorsDict}

    def marginalized_contraction(self, atomList):
        marginalizationDict = {atomKey + "_marg": cc.CoordinateCore(np.ones(shape=(2)), [atomKey]) for atomKey in
                               self.atoms if atomKey not in atomList}
        margContractor = coc.CoreContractor({**self.all_cores(), **marginalizationDict}, openColors=atomList)
        return margContractor.contract()

    def contract_partition(self):
        if self.headType != "expFactor":
            print("Warning: Partition of Tensor Model computed, but headtype expFactor!")
        return self.marginalized_contraction([]).values

    def evidence_contraction(self, evidenceDict):
        inferedFormulaTensorDict = {formulaKey: self.formulaTensorsDict[formulaKey].infer_on_evidenceDict(evidenceDict)
                                    for formulaKey in self.formulaTensorsDict}
        resContractor = coc.CoreContractor(inferedFormulaTensorDict, openColors=[atomKey for atomKey in self.atoms if
                                                                                 atomKey not in evidenceDict])
        return resContractor.contract()

    def visualize(self, evidenceDict={}, strengthMultiplier=4, strengthCutoff=10, fontsize=10, showFormula=True,
                  pos=None):
        return mv.visualize_model(self.expressionsDict,
                                  strengthMultiplier=strengthMultiplier,
                                  strengthCutoff=strengthCutoff,
                                  fontsize=fontsize,
                                  showFormula=showFormula,
                                  evidenceDict=evidenceDict,
                                  pos=pos)


if __name__ == "__main__":
    learnedFormulaDict = {
        "f0": ["A1", 1],
        "f1": [["not", ["A2", "and", "A3"]], 1.2],
        "f2": ["A2", 2.1]
    }
    tRep = TensorRepresentation(learnedFormulaDict)

    tRep.visualize(evidenceDict={"A2": 1, "A3": False})
    print(tRep.evidence_contraction({"A2": 1, "A3": False}).values)
    print(tRep.marginalized_contraction(["A2", "A3"]).values)

    print(tRep.contract_partition())



import tnreason.logic.coordinate_calculus as cc
import tnreason.logic.expression_generation as eg
import tnreason.logic.expression_utils as eu

import tnreason.representation.factdf_to_cores as ftoc
import tnreason.representation.sampledf_to_cores as stoc
import tnreason.representation.pairdf_to_cores as ptoc

import tnreason.optimization.generalized_als as gals

import numpy as np


class OptimizerBase:
    def __init__(self, skeletonExpression, candidatesDict=None):
        self.skeleton = skeletonExpression
        self.skeletonAtoms = eu.get_variables(skeletonExpression)
        self.candidatesDict = candidatesDict

        self.targetCore = None
        self.filterCore = None

    def generate_target_and_filterCore_from_exampleCores(self, positiveCore, negativeCore=None):
        if negativeCore is None:
            negativeCore = positiveCore.create_constant([], zero=True)
        self.targetCore = positiveCore.clone()
        self.filterCore = positiveCore.compute_or(negativeCore)

    def set_importance(self, importanceValues=None, importanceCore=None):
        if importanceCore is not None:
            self.filterCore = importanceCore
        elif importanceValues is not None:
            assert self.targetCore is not None, "Attempting to set ImportanceCore via values, but TargetCore has not yet been initialized."
            assert self.targetCore.values.shape == importanceValues.shape, "Shape of ImportanceCore does not match the TargetCore"
            self.filterCore = self.targetCore.clone()
            self.filterCore.values = importanceValues
        else:
            raise "FilterCore has not been initialized!"

    def balance_importance(self, positiveCore=None, negativeCore=None, strategy="pn-equality"):
        ## Compute number of positive and negative demonstration
        if positiveCore is None:
            positiveCore = self.targetCore.compute_and(self.filterCore)
        if negativeCore is None:
            negativeCore = self.targetCore.negate().compute_and(self.filterCore)
        posExNum = np.count_nonzero(positiveCore.values)
        negExNum = np.count_nonzero(negativeCore.values)

        ## Balance contributions to risk by positive and negative demonstration
        if strategy == "pn-equality":
            if posExNum == 0:
                posFactor = 1
                print("WARNING: No positive demonstration have been provided and loss is not balanced!")
            elif negExNum == 0:
                posFactor = 1
                print("WARNING: No negative demonstration have been provided and loss is not balanced!")
            else:
                posFactor = np.sqrt(negExNum / posExNum)
        impValues = posFactor * positiveCore.values + negativeCore.values
        self.set_importance(importanceValues=impValues)

    def als(self, sweepnum, verbose=False):
        self.optimizer = gals.GeneralizedALS(self.variablesCoresDict, self.fixedCoresDict)
        self.optimizer.set_targetCore(self.targetCore)
        self.optimizer.set_filterCore(self.filterCore)

        self.optimizer.sweep(sweepnum=sweepnum, contractionScheme=self.skeleton, verbose=verbose)
        self.variablesCoresDict = self.optimizer.variableCoresDict

    def get_solution(self, adjustVariablesCoresDict=True):
        self.solutionDict = {}
        for legKey in self.variablesCoresDict:
            maxPos = np.argmax([abs(val) for val in self.variablesCoresDict[legKey].values])
            self.solutionDict[legKey] = self.candidatesDict[legKey][maxPos]
            if adjustVariablesCoresDict:
                self.variablesCoresDict[legKey].values = np.zeros(shape=self.variablesCoresDict[legKey].values.shape)
                self.variablesCoresDict[legKey].values[maxPos] = 1

        self.solutionExpression = eg.replace_atoms(self.skeleton, self.solutionDict)


class SampleBasedOptimizer(OptimizerBase):
    ## Initializes the fixedCoresDict using the sampleDf
    def generate_fixedCores_sampleDf(self, sampleDf):
        self.fixedCoresDict = {
            atomKey: cc.CoordinateCore(stoc.sampleDf_to_universal_core(sampleDf, self.candidatesDict[atomKey]),
                                       ["j", atomKey]) for atomKey in self.skeletonAtoms
        }

    ## Initializes all VariableCores (shapes follow from fixedCoresDict) with independent coordinates drawn uniformly from [0,1)
    def random_initialize_variableCoresDict(self):
        self.variablesCoresDict = {}
        for legKey in self.fixedCoresDict:
            varColors = [c for c in self.fixedCoresDict[legKey].colors if c != "j"]
            varDims = [self.fixedCoresDict[legKey].values.shape[i] for i, c in
                       enumerate(self.fixedCoresDict[legKey].colors) if c != "j"]
            self.variablesCoresDict[legKey] = cc.CoordinateCore(np.random.random(varDims), varColors)


## NOT MAINTAINED FROM HERE ! ##
## TO BE IMPLEMENTED: VariableBasedOptimizer(OptimizerBase)
class AtomicLearner(OptimizerBase):

    def generate_fixedCores_sampleDf(self, sampleDf, candidatesDict=None):
        if candidatesDict is not None:
            self.candidatesDict = candidatesDict

        self.fixedCoresDict = {}

        for atomKey in self.skeletonAtoms:
            values = stoc.sampleDf_to_universal_core(sampleDf, self.candidatesDict[atomKey])
            self.fixedCoresDict[atomKey] = cc.CoordinateCore(values, ["j", atomKey])

    def random_initialize_variableCoresDict(self):
        self.variablesCoresDict = {}
        for legKey in self.fixedCoresDict:
            varColors, varDims = atomic_find_var(self.fixedCoresDict[legKey])
            self.variablesCoresDict[legKey] = cc.CoordinateCore(np.random.random(varDims), varColors)

    def set_targetCore(self, targetCore=None, length=0, targetIsFilter=True):
        if targetCore is None:
            self.targetCore = cc.CoordinateCore(np.ones([length]), ["j"])
        else:
            self.targetCore = targetCore

        ## If targetIsFilter, only positive demonstration are considered in training, negative are ignored
        if targetIsFilter:
            self.filterCore = self.targetCore.clone()


def atomic_find_var(fixedCore):
    varColors = []
    varDims = []
    for i, color in enumerate(fixedCore.colors):
        if color != "j":
            varColors.append(color)
            varDims.append(fixedCore.values.shape[i])
    return varColors, varDims


class VariableLearner(OptimizerBase):

    def generate_fixedCores_sampledf(self, sampleDf):
        self.candidatesDict = {}
        self.fixedCoresDict = {}

        for atomKey in self.skeletonAtoms:
            if "," in atomKey:
                indSpec = atomKey.split("(")[1][:-1]
                indKey1, indKey2 = indSpec.split(",")
                relationKey = atomKey.split("(")[0]

                coreValues, candidates, latency = stoc.sampleDf_to_relation_values(sampleDf, indKey1, indKey2)

                self.candidatesDict[atomKey] = candidates
                self.fixedCoresDict[atomKey] = cc.CoordinateCore(coreValues, [indKey1, relationKey, indKey2])
            else:
                indKey = atomKey.split("(")[1][:-1]
                classKey = atomKey.split("(")[0]

                coreValues, candidates, latency = stoc.sampleDf_to_class_values(sampleDf, indKey)

                self.candidatesDict[atomKey] = candidates
                self.fixedCoresDict[atomKey] = cc.CoordinateCore(coreValues, [indKey, classKey])

    def generate_fixedCores_factDf(self, df, individualsDict, candidatesDict, prefix=""):
        self.candidatesDict = candidatesDict
        self.fixedCoresDict = {}

        for atomKey in self.skeletonAtoms:
            if "," in atomKey:
                indSpec = atomKey.split("(")[1][:-1]
                indKey1, indKey2 = indSpec.split(",")
                relationKey = atomKey.split("(")[0]
                if not relationKey.startswith("R"):
                    relationKey = "R_" + relationKey
                coreValues, latency = ftoc.factDf_to_relation_values(df,
                                                                     individuals1=individualsDict[indKey1],
                                                                     individuals2=individualsDict[indKey2],
                                                                     relations=candidatesDict[atomKey],
                                                                     prefix=prefix)
                self.fixedCoresDict[atomKey] = cc.CoordinateCore(coreValues, [indKey1, relationKey, indKey2])

            else:
                indKey = atomKey.split("(")[1][:-1]
                classKey = atomKey.split("(")[0]
                if not classKey.startswith("C"):
                    classKey = "C_" + classKey
                coreValues, latency = ftoc.factDf_to_class_values(df,
                                                                  individuals=individualsDict[indKey],
                                                                  classes=candidatesDict[atomKey],
                                                                  prefix=prefix)
                self.fixedCoresDict[atomKey] = cc.CoordinateCore(coreValues, [indKey, classKey])

    def generate_targetCore_pairDf(self, pairDf, individualsDict, targetIsFilter=True):
        skeletonIndividuals = eu.get_individuals(self.skeleton)
        coreValues, latency = ptoc.pairDf_to_target_values(pairDf, individualsDict, skeletonIndividuals)
        self.targetCore = cc.CoordinateCore(coreValues, skeletonIndividuals)
        if targetIsFilter:
            self.filterCore = self.targetCore.clone()

    def random_initialize_variableCoresDict(self):
        self.variablesCoresDict = {}
        for legKey in self.fixedCoresDict:
            varColor, varDim = variable_find_var(self.fixedCoresDict[legKey])
            self.variablesCoresDict[legKey] = cc.CoordinateCore(np.random.random(varDim), [varColor])


def variable_find_var(fixedCore):
    varColors = []
    for color in fixedCore.colors:
        if color.startswith("C") or color.startswith("R"):
            varColors.append(color)
    if len(varColors) != 1:
        raise TypeError("Variables in Core with colors {} not found correctly!".format(fixedCore.colors))
    return varColors[0], fixedCore.values.shape[fixedCore.colors.index(varColors[0])]


if __name__ == "__main__":
    import pandas as pd

    samDf = pd.read_csv("./demonstration/generation/synthetic_test_data/generated_sampleDf.csv").astype("int64")

    skeleton = ["P0", "and", "P1"]
    candidatesDict = {
        "P0": list(samDf.columns),
        "P1": list(samDf.columns)
    }
    learner = AtomicLearner(skeleton)
    learner.generate_fixedCores_sampleDf(samDf, candidatesDict)

    learner.random_initialize_variableCoresDict()

    learner.set_targetCore(length=samDf.values.shape[0])
    learner.set_importance(importanceCore=cc.CoordinateCore(learner.targetCore.values * 10, learner.targetCore.colors))
    learner.balance_importance()
    print(learner.filterCore.values)
    print(learner.targetCore.values)
    learner.als(2)

    learner.get_solution()
    # print(learner.solutionExpression)



import networkx as nx

import os
from matplotlib import pyplot as plt
from matplotlib import animation as animation

from tnreason.logic import expression_utils as eu


def visualize_model(expressionsDict,
                    strengthMultiplier=4,
                    strengthCutoff=10,
                    fontsize=10,
                    showFormula=False,
                    evidenceDict={},
                    pos=None,
                    savePath=None,
                    show=True,
                    title="Visualization of the MLN"):
    expressionsList = [expressionsDict[key][0] for key in expressionsDict]
    atomsList = eu.get_all_variables(expressionsList)

    ## Collect edges for position optimization
    edges = []
    for expressionKey in expressionsDict:
        for atom in eu.get_variables(expressionsDict[expressionKey][0]):
            edges.append([atom, expressionKey])
    graph = nx.Graph()
    graph.add_nodes_from(atomsList)
    graph.add_nodes_from(expressionsDict.keys())
    graph.add_edges_from(edges)
    if pos is None:
        pos = nx.spring_layout(graph, k=0.8)

    ## Draw Nodes
    trueColor = "blue"
    falseColor = "red"

    # Known Trues
    atomFontSize = fontsize * 100
    nx.draw_networkx_nodes(graph, pos,
                           nodelist=[atomKey for atomKey in evidenceDict if bool(evidenceDict[atomKey])],
                           node_color=trueColor,
                           node_size=atomFontSize,
                           alpha=0.6)
    # Known False
    nx.draw_networkx_nodes(graph, pos,
                           nodelist=[atomKey for atomKey in evidenceDict if not bool(evidenceDict[atomKey])],
                           node_color=falseColor,
                           node_size=atomFontSize,
                           alpha=0.6)

    # Unknown Atoms
    nx.draw_networkx_nodes(graph, pos,
                           nodelist=[atomKey for atomKey in atomsList if atomKey not in evidenceDict.keys()],
                           node_color="grey",
                           node_size=atomFontSize,
                           alpha=0.2)
    nx.draw_networkx_labels(graph, pos, {atomKey: atomKey for atomKey in atomsList}, font_size=fontsize)

    if showFormula:
        expressionLabels = {expressionKey: expressionsDict[expressionKey][0] for expressionKey in expressionsDict}
    else:
        expressionLabels = {expressionKey: expressionKey for expressionKey in expressionsDict}
    nx.draw_networkx_labels(graph, pos, expressionLabels, font_size=fontsize)

    ## Draw Edges
    colorList = ["r", "g", "b", "r", "g", "b", "r", "g", "b"]
    for i, expressionKey in enumerate(expressionsDict.keys()):
        strength = min(expressionsDict[expressionKey][1], strengthCutoff)
        drawEdges = []
        for atom in eu.get_variables(expressionsDict[expressionKey][0]):
            drawEdges.append([atom, expressionKey])
        nx.draw_networkx_edges(graph, pos,
                               edgelist=drawEdges,
                               width=strengthMultiplier * strength,
                               alpha=0.2,
                               edge_color=colorList[i])
    plt.title(title, fontsize=15)
    if savePath is not None:
        plt.savefig(savePath)
    if show:
        plt.show()
    return pos


def create_animation(pngDirPath, savePath):
    images = []
    for filename in os.listdir(pngDirPath):
        if filename.endswith('.png'):
            images.append(plt.imread(os.path.join(pngDirPath, filename)))
    # Create an animation
    fig = plt.figure()
    im = plt.imshow(images[0])
    ani = animation.FuncAnimation(fig, lambda i: im.set_array(images[i]), frames=len(images))

    # Save the animation as an MP4 file
    ani.save(savePath, writer='ffmpeg', fps=0.5)


if __name__ == "__main__":
    #create_animation("./demonstration/visualizations/",savePath="./animation.mp4")

    exDict = {
        "e0": ["a2", 1],
        "e1": [[["a2", "and", ["not", "a3"]], "and", ["a6", "and", ["not", "a7"]]], 5],
        "e2": [["a4", "and", ["not", "a2"]], 2],
        "e4": ["a5", 1],
        "e5": ["a5", 2]
    }

    visualize_model(exDict, evidenceDict={"a2": 1, "a3": 0})

    exit()
    import matplotlib.pyplot as plt
    import matplotlib.cm as cm
    import matplotlib.animation as animation

    img = []  # some array of images
    frames = []  # for storing the generated images
    fig = plt.figure()
    for i in range(6):
        frames.append([plt.imshow(img[i], cmap=cm.Greys_r, animated=True)])

    ani = animation.ArtistAnimation(fig, frames, interval=50, blit=True,
                                    repeat_delay=1000)
    # ani.save('movie.mp4')
    plt.show()






    G = nx.Graph()
    G.add_edges_from([(12, 2)])  # , (2, 3), (3, 4), (4, 1)])

    pos = {12: [10, 0], 2: [1, 0]}
    nx.draw(G, pos=pos)
    plt.show()



import numpy as np
import time

import tnreason.representation.ttl_to_csv as ttodf


def generate_factDf(ttlPath, limit=None):
    return ttodf.generate_csv_list(ttlPath, limit)


def filter_df(df, column, values):
    startTime = time.time()
    df = df[df[column].isin(values)]
    endTime = time.time()
    return df, endTime - startTime

def prefixify(lis, prefix):
    return np.array([prefix+entry.split("(")[0] for entry in lis])


def factDf_to_class_values(df, individuals=None, classes=None, prefix=""):
    startTime = time.time()

    if classes is not None:
        classes = prefixify(classes,prefix)

    classes = np.array(classes)
    individuals = np.array(individuals)

    classDf = df[df["predicate"].isin(["http://www.w3.org/1999/02/22-rdf-syntax-ns#type", "rdf:type"])][
        ["subject", "object"]]

    if classes is None:
        classes = np.unique(df["object"].values)
    if individuals is not None:
        classDf, latency = filter_df(classDf, "subject", individuals)
    else:
        individuals = np.unique(classDf["subject"].values)

    repTensor = np.zeros((len(individuals), len(classes)))
    for i, row in classDf.iterrows():
        subPos = np.argwhere(individuals == row["subject"])
        obPos = np.argwhere(classes == row["object"])
        repTensor[subPos, obPos] = 1
    endTime = time.time()

    return repTensor, endTime - startTime


def factDf_to_relation_values(df, individuals1=None, individuals2=None, relations=None, prefix=""):
    startTime = time.time()

    if prefix != "":
        if relations is not None:
            relations = prefixify(relations,prefix)

    if individuals1 is not None:
        df, latency = filter_df(df, "subject", individuals1)
    else:
        individuals1 = np.unique(df["subject"].values)

    if individuals2 is not None:
        df, latency = filter_df(df, "object", individuals2)
    else:
        individuals2 = np.unique(df["object"].values)

    if relations is not None:
        df, latency = filter_df(df, "predicate", relations)
    else:
        relations = np.unique(df["predicate"].values)

    relations = np.array(relations)
    individuals1 = np.array(individuals1)
    individuals2 = np.array(individuals2)

    repTensor = np.zeros((len(individuals1), len(relations), len(individuals2)))
    for i, row in df.iterrows():
        subPos = np.argwhere(individuals1 == row["subject"])
        predPos = np.argwhere(relations == row["predicate"])
        obPos = np.argwhere(individuals2 == row["object"])
        repTensor[subPos, predPos, obPos] = 1
    endTime = time.time()

    return repTensor, endTime - startTime



from tnreason.logic import expression_generation as eg
# from tnreason.logic import expression_calculus as ec
from tnreason.logic import expression_utils as eu

from pracmln import MLN, Database
from pracmln.mlnlearn import MLNLearn
from pracmln.utils import config, locs
from pracmln.utils.project import PRACMLNConfig
from pracmln.utils.config import global_config_filename
import os


class WeightEstimator:
    def __init__(self, solutionList=None):
        self.create_mln(solutionList)
        self.db = Database(self.mln)

    def create_mln(self, solutionList):
        self.mln = MLN(grammar="StandardGrammar",
                       logic="FirstOrderLogic")
        variablesList = eu.get_all_variables(solutionList)
        for variable in variablesList:
            self.mln << variable
        for expression in solutionList:
            self.mln << "0 " + eg.generate_pracmln_formulastring(expression)

    def extend_db(self, factDf):

        for i, row in factDf.iterrows():
            if row["predicate"] == "rdf:type":
                cla = row["object"]
                ind = row["subject"]
                self.db << cla + "(" + ind + ")"
            else:
                ind1 = row["subject"]
                ind2 = row["object"]
                rel = row["predicate"]
                self.db << rel + "(" + ind1 + "," + ind2 + ")"

    def learn_weights(self):
        DEFAULT_CONFIG = os.path.join(locs.user_data, global_config_filename)
        conf = PRACMLNConfig(DEFAULT_CONFIG)

        learner = MLNLearn(conf, mln=self.mln, db=self.db)
        result = learn.run()


if __name__ == "__main__":
    import pandas as pd

    factDf = pd.read_csv("./demonstration/generation/synthetic_test_data/generated_factDf.csv")

    # solutionList = ["bearbeitet(y,x)", "and", "enthaelt(x,z)"]
    # westimator = WeightEstimator(solutionList)



from tnreason.logic import expression_utils as eu

from tnreason.contraction import core_contractor as coc
from tnreason.contraction import expression_evaluation as ee

from tnreason.model import logic_model as lm

import numpy as np
import pandas as pd


## Shifted main sampling functionality to sampling

class TensorMLN:
    def __init__(self, expressionsDict, formulaCoreDict=None):
        self.expressionsDict = expressionsDict
        self.atomKeys = eu.get_all_variables([self.expressionsDict[formulaKey][0] for formulaKey in self.expressionsDict])
        self.formulaCoreDict = formulaCoreDict

    def infer_on_evidenceDict(self, evidenceDict={}):
        inferedExpressionsDict = {}
        for key in self.expressionsDict:
            inferedFormula = lm.infer_expression(self.expressionsDict[key][0], evidenceDict)
            if inferedFormula not in ["Thing", "Nothing"]:
                inferedFormula = lm.reduce_double_not(inferedFormula)
                inferedExpressionsDict[key] = [inferedFormula, self.expressionsDict[key][1]]
        return TensorMLN(inferedExpressionsDict)

    def reduce_double_formulas(self):
        checkedKeys = []
        reducedExpressionDict = {}
        for key in self.expressionsDict:
            if key not in checkedKeys:
                checkedKeys.append(key)
                keyFormula, keyWeight = self.expressionsDict[key]
                for otherKey in self.expressionsDict:
                    if otherKey not in checkedKeys and lm.equality_check(keyFormula, self.expressionsDict[otherKey][0]):
                        checkedKeys.append(otherKey)
                        keyWeight = keyWeight + self.expressionsDict[otherKey][1]
                reducedExpressionDict[key] = [keyFormula, keyWeight]
        self.expressionsDict = reducedExpressionDict

    def initialize_formulaCoreDict(self):
        self.formulaCoreDict = {
            formulaKey: ee.ExpressionEvaluator(self.expressionsDict[formulaKey][0],
                                                initializeBasisCores=True).create_formula_factor().weighted_exponentiation(
                self.expressionsDict[formulaKey][1])
            for formulaKey in self.expressionsDict}

    def compute_marginalized(self, marginalKeys, contractionMethod="formulaDict", optimizationMethod="GreedyHeuristic"):
        if contractionMethod == "formulaDict":
            if self.formulaCoreDict is None:
                self.initialize_formulaCoreDict()
            contractionDict = self.formulaCoreDict.copy()
        elif contractionMethod == "basisCalculus":
            contractionDict = {}
            for expression in self.expressionsDict:
                pass
        else:
            raise ValueError("Contraction Method {} not understood!".format(contractionMethod))
        contractor = coc.CoreContractor(contractionDict, openColors=marginalKeys)
        return contractor.contract(optimizationMethod=optimizationMethod).normalize()

    ## To be implemented: Here we need Tensor Network contractions
    def create_independent_atom_sample(self, atomSampleKey):
        marginalProbCore = self.compute_marginalized([atomSampleKey])
        return np.random.multinomial(1, marginalProbCore.values)[0] == 0

    def create_independent_sample(self):
        return {atomKey: self.create_independent_atom_sample(atomKey) for atomKey in self.atomKeys}

    def gibbs(self, repetitionNum=10, verbose=True):
        sampleDict = self.create_independent_sample()
        for repetitionPos in range(repetitionNum):
            if verbose:
                print("## Gibbs Iteration {} ##".format(repetitionPos))
            for refinementAtomKey in sampleDict:
                samplerMLN = self.infer_on_evidenceDict(
                    {key: sampleDict[key] for key in sampleDict if key != refinementAtomKey})
                samplerMLN.reduce_double_formulas()
                if refinementAtomKey not in samplerMLN.atomKeys:
                    print("Warning: Infered MLN is empty on key {}".format(refinementAtomKey))
                    samplerMLN = TensorMLN({refinementAtomKey: [str(refinementAtomKey), 0]})
                sampleDict[refinementAtomKey] = samplerMLN.create_independent_atom_sample(refinementAtomKey)
            if verbose:
                print("SampleDict is {}".format(sampleDict))
        return sampleDict

    def generate_sampleDf(self, sampleNum, method="Gibbs10"):
        df = pd.DataFrame(columns=self.atomKeys)
        for ind in range(sampleNum):
            if method.startswith("Gibbs"):
                repetitionNum = int(method.split("Gibbs")[1])
                row_df = pd.DataFrame(self.gibbs(repetitionNum=repetitionNum, verbose=False), index=[ind])
                df = pd.concat([df, row_df])
        return df.astype("int64")


#def create_formulaCoreDict(expressionsDict):
#    return {formulaKey: ec.calculate_expressionCore(expressionsDict[formulaKey][0]).weighted_exponentiation(
#        expressionsDict[formulaKey][1])
#        for formulaKey in expressionsDict}





if __name__ == "__main__":
    example_expression_dict = {
        "e0": [["not", ["Unterschrank(z)", "and", ["not", "Moebel(z)"]]], 20],
        "e0.5": ["Moebel(z)", 4],
        "e0.625": ["Sledz", 4],
        "e0.626": [["not", ["Moebel(z)", "and", "Sledz"]], 5],
        "e0.75": [["not", [["Unterschrank(z)", "and", "Sledz"], "and", ["not", "Ausgangsrechnung(x)"]]], 2],
        "e1": [["not", "Ausgangsrechnung(x)"], 12],
        "e2": [[["not", "Ausgangsrechnung(x)"], "and", ["not", "Rechnung(x)"]], 14]
    }
    ## 1 = True, 0 = False
    example_evidence_dict = {
        "Unterschrank(z)": 1,
        "Ausgangsrechnung(x)": 1
    }

    tn_mln = TensorMLN(example_expression_dict)
    print(tn_mln.generate_sampleDf(int(100)))

    infered_mln = tn_mln.infer_on_evidenceDict(example_evidence_dict)



import tnreason.contraction.bc_contraction_generation as bcg
import tnreason.contraction.core_contractor as coc

import tnreason.model.formula_tensors as ft
import tnreason.model.tensor_model as tm
import tnreason.model.entropies as ent

import tnreason.logic.coordinate_calculus as cc
import tnreason.logic.expression_generation as eg

import numpy as np


class MLEBase:
    def __init__(self, skeletonExpression, candidatesDict, variableCoresDict=None, learnedFormulaDict={},
                 sampleDf=None):
        if variableCoresDict is None:
            variableCoresDict = {
                key: cc.CoordinateCore(np.random.random(size=len(candidatesDict[key])), [key], key + "_varCore") for key
                in candidatesDict}
        self.superposedFormulaTensor = ft.SuperposedFormulaTensor(skeletonExpression, candidatesDict,
                                                                  parameterCoresDict=variableCoresDict)
        if sampleDf is not None:
            self.load_sampleDf(sampleDf)

        self.learnedFormulaDict = learnedFormulaDict
        self.fixedFormulaTensors = tm.TensorRepresentation(learnedFormulaDict, headType="weightedTruthEvaluation")
        self.compute_fixedCores_likelihoodCorrection()

        self.skeletonExpression = skeletonExpression
        self.candidatesDict = candidatesDict
        self.candidatesAtomsList = []
        for candidatesKey in self.candidatesDict:
            for atom in self.candidatesDict[candidatesKey]:
                if atom not in self.candidatesAtomsList:
                    self.candidatesAtomsList.append(atom)

    def load_sampleDf(self, sampleDf):
        self.dataTensor = ft.DataTensor(sampleDf)

    def create_exponentiated_variables(self):
        ## Creates exponentiated factor to the variables tbo
        self.variablesExpFactor = coc.CoreContractor(self.superposedFormulaTensor.get_cores(),
                                                     openColors=self.candidatesAtomsList).contract(
            optimizationMethod="GreedyHeuristic").exponentiate()

        ## To do: Think of removing this bottleneck: Constract the current variables to a single core instead of exp TN!

    def contract_partition_gradient(self, tboVariableKey):
        self.create_exponentiated_variables()
        contractionDict = {**self.fixedFormulaTensors.all_cores(),
                           "variablesExpFactor": self.variablesExpFactor,
                           **self.superposedFormulaTensor.get_cores(
                               parameterExceptionKeys=[tboVariableKey])}
        return coc.CoreContractor(contractionDict, openColors=self.superposedFormulaTensor.parameterCoresDict[
            tboVariableKey].colors).contract(optimizationMethod="GreedyHeuristic").multiply(
            1 / self.contract_partition(reDoExpVariables=False))

    def contract_partition(self, visualize=False, reDoExpVariables=True):
        if reDoExpVariables:
            self.create_exponentiated_variables()
        contractor = coc.CoreContractor({  # **self.fixedFormulaTensors.all_cores(),
            **self.fixedFormulaTensors.get_cores(headType="expFactor"),
            "variablesExpFactor": self.variablesExpFactor})
        if visualize:
            contractor.visualize(title="Partition Function")
        return contractor.contract().values

    def contract_data_gradient(self, tboVariableKey):
        contractor = coc.CoreContractor(
            {**self.superposedFormulaTensor.get_cores(parameterExceptionKeys=[tboVariableKey]),
             **self.dataTensor.get_cores()
             },
            openColors=self.superposedFormulaTensor.parameterCoresDict[
                tboVariableKey].colors)
        return contractor.contract().multiply(1 / self.dataTensor.dataNum)

    ## Compute Estimation Metric: Log likelihood (= negative empirical cross entropy)
    def compute_likelihood(self, visualize=False):
        contractor = coc.CoreContractor({**self.superposedFormulaTensor.get_cores(),
                                         **self.dataTensor.get_cores()})
        if visualize:
            contractor.visualize(title="Likelihood VariableCores")

        return contractor.contract().values / (self.dataTensor.dataNum) - np.log(
            self.contract_partition()) + self.formulaCorrectionTerm

    def compute_fixedCores_likelihoodCorrection(self):
        self.formulaCorrectionTerm = np.sum(
            [coc.CoreContractor({**self.fixedFormulaTensors.get_cores([formulaKey], headType="weightedTruthEvaluation"),
                                 **self.dataTensor.get_cores()}).contract().values
             for formulaKey in self.learnedFormulaDict]) / self.dataTensor.dataNum

    def compute_empirical_KL_divergence(self, sampleDf):
        return -self.compute_likelihood() - ent.empirical_shannon_entropy(sampleDf, self.candidatesAtomsList)

    ## Optimization preparation
    def random_initialize_variableCoresDict(self):
        self.superposedFormulaTensor.random_initialize_parameterCoresDict()

    ## Output of the solution
    def get_solution_expression(self):
        return eg.replace_atoms(self.skeletonExpression,
                                self.superposedFormulaTensor.get_largest_weight_as_solutionMap())


class GradientDescentMLE(MLEBase):

    def compute_gradient(self, tboVariableKey):
        partition_gradient = self.contract_partition_gradient(tboVariableKey)
        data_gradient = self.contract_data_gradient(tboVariableKey)
        return partition_gradient.sum_with(data_gradient.multiply(-1))

    def descent_step(self, tboVariableKey, stepWidth=1):
        self.superposedFormulaTensor.parameterCoresDict[tboVariableKey] = \
            self.superposedFormulaTensor.parameterCoresDict[tboVariableKey].sum_with(
                self.compute_gradient(tboVariableKey).multiply(-1 * stepWidth)
            )

    def alternating_gradient_descent(self, sweepNum, stepWidth, computeLikelihood=True, verbose=True):
        safeLikeLihood = -2.2250738585072014e308  # The smallest float
        likelihoods = np.empty(shape=(sweepNum, len(self.superposedFormulaTensor.parameterCoresDict)))
        for sweep in range(sweepNum):
            for i, variableKey in enumerate(self.superposedFormulaTensor.parameterCoresDict):
                self.descent_step(variableKey, stepWidth)
                if computeLikelihood:
                    likelihoods[sweep, i] = self.compute_likelihood()
                    if likelihoods[sweep, i] < safeLikeLihood and verbose:
                        print(
                            "Warning: Likelihood increased on variableCore {} in sweep {}.".format(variableKey, sweep))
                    if likelihoods[sweep, i] > 0:
                        print("Warning: Positive Likelihood on variableCore {} in sweep {}.".format(variableKey, sweep))
                    safeLikeLihood = likelihoods[sweep, i]
        return likelihoods


## Does not yet make use of the superposedFormulaTensor -> Still on pure conjunctions
class AlternatingNewtonMLE(MLEBase):
    def create_atom_selectors(self):
        ## Creates atom selector cores selecting atom activation
        self.atomSelectorDict = {}
        for placeHolderKey in self.candidatesDict:
            for i, atomKey in enumerate(self.candidatesDict[placeHolderKey]):
                coreValues = np.ones(shape=(len(self.candidatesDict[placeHolderKey]), 2))
                coreValues[i, 0] = 0
                self.atomSelectorDict[placeHolderKey + "_enumeratedWorlds_" + atomKey] = cc.CoordinateCore(
                    coreValues, [placeHolderKey, atomKey], placeHolderKey + "_enumeratedWorlds_" + atomKey)

    def create_variable_gradient(self, tboVariableKey):
        return {key: self.variableCoresDict[key] for key in self.variableCoresDict if key != tboVariableKey}

    def contract_double_gradient_exponential(self, tboVariableKey):

        self.formulaCoreDict = bcg.generate_formulaCoreDict(self.learnedFormulaDict)

        contractionDict = {**self.formulaCoreDict, "variablesExpFactor": self.variablesExpFactor,
                           **self.create_variable_gradient(tboVariableKey),
                           **self.atomSelectorDict,
                           **copy_CoreDict(self.create_variable_gradient(tboVariableKey), suffix="out"),
                           **copy_CoreDict(self.atomSelectorDict, suffix="out",
                                           exceptionColors=self.candidatesAtomsList),
                           }
        openColors = self.variableCoresDict[tboVariableKey].colors
        openColors = openColors + [color + "_out" for color in openColors]
        contractor = coc.CoreContractor(contractionDict, openColors=openColors)
        return contractor.contract(optimizationMethod="GreedyHeuristic")

    ## Algorithm
    def alternating_newton(self, sweepNum=10, dampFactor=0.001, monotoneousCondition=True, regParameter=0,
                           verbose=True):
        self.create_atom_selectors()
        self.variableCoresDict = self.superposedFormulaTensor.parameterCoresDict

        self.likelihood = self.compute_likelihood()
        likelihoods = np.empty(shape=(sweepNum, len(self.variableCoresDict)))
        for sweep in range(sweepNum):
            print("### SWEEP {} ###".format(sweep))
            for i, variableKey in enumerate(self.variableCoresDict):
                self.newton_step(variableKey, dampFactor=dampFactor, monotoneusCondition=monotoneousCondition,
                                 regParameter=regParameter, verbose=verbose)
                likelihoods[sweep, i] = self.likelihood
        return likelihoods

    def newton_step(self, tboVariableKey, dampFactor, monotoneusCondition, regParameter, verbose=True):
        expGradient = self.contract_partition_gradient(tboVariableKey)

        vector = expGradient.sum_with(
            self.contract_data_gradient(tboVariableKey).multiply(-1 / self.contract_partition()))

        ## Operator
        doubleExpGradient = self.contract_double_gradient_exponential(tboVariableKey).multiply(-1)

        empOperator = expGradient.clone()
        empOperator.colors = [color + "_out" for color in empOperator.colors]
        empOperator = empOperator.compute_and(self.contract_data_gradient(tboVariableKey))

        operator = empOperator.sum_with(doubleExpGradient)

        ## Flatten
        outColors = [color for color in operator.colors if color.endswith("_out")]
        inColors = [color for color in operator.colors if color not in outColors]

        outshape = [operator.values.shape[i] for i, color in enumerate(operator.colors) if color in outColors]
        inshape = [operator.values.shape[i] for i, color in enumerate(operator.colors) if color in inColors]

        outDim = np.prod(outshape)
        inDim = np.prod(inshape)

        operator.reorder_colors(inColors + outColors)
        vector.reorder_colors(inColors)

        flattenedOperatorValues = operator.values.reshape(inDim, outDim) + regParameter * np.eye(inDim)
        flattenedVectorValues = vector.values.reshape(inDim)

        ## Solve Newton
        update, residuals, rank, singular_values = np.linalg.lstsq(flattenedOperatorValues, flattenedVectorValues)

        update.reshape(inshape)
        updateCore = cc.CoordinateCore(update.reshape(inshape), inColors)

        oldCore = self.variableCoresDict[tboVariableKey].clone()

        self.variableCoresDict[tboVariableKey] = self.variableCoresDict[tboVariableKey].sum_with(
            updateCore.multiply(dampFactor))

        if monotoneusCondition:
            likelihood = self.compute_likelihood()
            if likelihood < self.likelihood:
                print("Update of {} is not accepted.".format(tboVariableKey))
                self.variableCoresDict[tboVariableKey] = oldCore
            else:
                self.likelihood = likelihood

        if verbose:
            print("Update Norm", np.linalg.norm(update), "Operator Condition", np.linalg.cond(flattenedOperatorValues))
            print(self.compute_likelihood())


def copy_CoreDict(tbdDict, suffix="", exceptionColors=[]):
    doubleDict = {}
    for key in tbdDict:
        oldCore = tbdDict[key]
        newColors = []
        for color in oldCore.colors:
            if color in exceptionColors:
                newColors.append(color)
            else:
                newColors.append(color + "_" + suffix)
        doubleDict[key + "_" + suffix] = cc.CoordinateCore(oldCore.values, newColors)
    return doubleDict


if __name__ == "__main__":
    from tnreason.logic import coordinate_calculus as cc

    atomDict = {
        "a": cc.CoordinateCore(np.random.binomial(n=1, p=0.8, size=(10, 7, 5)), ["l1", "y", "z"], name="a"),
        "b": cc.CoordinateCore(np.random.binomial(n=1, p=0.8, size=(10, 7, 5)), ["l2", "q", "z"], name="b"),
        "c": cc.CoordinateCore(np.random.binomial(n=1, p=0.4, size=(10, 7, 5)), ["l3", "q", "z"], name="c"),
    }

    skeletonExpression = ["P1", "and", "P2"]
    candidatesDict = {"P1": list(atomDict.keys()),
                      "P2": list(atomDict.keys()),
                      }

    variableCoresDict = {
        "v1": cc.CoordinateCore(np.zeros(shape=(3, 2)), ["P1", "H1"]),
        "v2": cc.CoordinateCore(np.zeros(shape=(3, 2)), ["P2", "H1"]),
    }

    learnedFormulaDict = {
        "f0": ["b", 10],
        "f1": [["not", ["a", "and", "b"]], 5],
        "f2": ["c", 2]
    }

    import tnreason.model.generate_test_data as gtd

    sampleDf = gtd.generate_sampleDf(learnedFormulaDict, 100)

    optimizer = AlternatingNewtonMLE(skeletonExpression, candidatesDict, variableCoresDict, {}, sampleDf=sampleDf)

    optimizer.random_initialize_variableCoresDict()

    optimizer.create_exponentiated_variables()

    optimizer.alternating_newton(10)



import numpy as np

alphabet = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
            'u', 'v', 'w', 'x', 'y', 'z']


class CoordinateCore:
    def __init__(self, core_values, core_colors, name=None):
        self.values = core_values
        self.colors = core_colors
        self.name = name

        if len(core_colors) != len(core_values.shape):
            raise TypeError("Number of Colors does not match the Value Shape in Core {}!".format(name))

        ## We allow multiple colors for now, although there might be problems
        #if len(core_colors) != len(set(core_colors)):
        #    raise TypeError("There are duplicate colors in Core {}!".format(name))

    def clone(self):
        return CoordinateCore(np.copy(self.values), self.colors.copy(), self.name)

    def negate(self, ignore_ones=False):
        newName = ["not", self.name]

        if ignore_ones:
            return CoordinateCore(np.copy(- self.values), self.colors.copy(), newName)
        else:
            return CoordinateCore(np.ones(self.values.shape) - np.copy(self.values), self.colors.copy(), newName)

    def multiply(self, factor):
        return CoordinateCore(factor*self.values, self.colors, self.name)

    def compute_and(self, in_core1):
        core0 = self.clone()
        core1 = in_core1.clone()

        colorDict = {core0.colors[i]: alphabet[i] for i in range(len(core0.colors))}
        i = len(core0.colors)
        for color in core1.colors:
            if color not in colorDict:
                colorDict[color] = alphabet[i]
                i += 1

        core0_string = "".join([colorDict[color] for color in core0.colors])
        core1_string = "".join([colorDict[color] for color in core1.colors])
        premise_string = ",".join([core0_string, core1_string])

        ## Modified to handle multiply colors in a core
        head_string = ""
        out_colors = []
        for color in core0.colors:
            if colorDict[color] not in head_string:
                head_string += colorDict[color]
                out_colors.append(color)

        for color in core1.colors:
            if colorDict[color] not in head_string:
                head_string += colorDict[color]
                out_colors.append(color)
        contraction_string = "->".join([premise_string, head_string])

        return CoordinateCore(np.einsum(contraction_string, core0.values, core1.values),
                              out_colors,
                              [core0.name, "and", core1.name])

    def reorder_colors(self, newColors):
        oldColors = self.colors.copy()
        oldValues = np.copy(self.values)

        colorDict = {oldColors[i]: alphabet[i] for i in range(len(oldColors))}
        old_string = "".join([colorDict[color] for color in oldColors])
        new_string = "".join([colorDict[color] for color in newColors])
        contraction_string = "->".join([old_string, new_string])

        newValues = np.einsum(contraction_string, oldValues)

        self.values = newValues
        self.colors = newColors

    def sum_with(self, core1):
        if core1.colors != self.colors:
            self.reorder_colors(core1.colors)
        if core1.values.shape != self.values.shape:
            raise TypeError("Shapes do not match for summations!")
        return CoordinateCore(self.values + core1.values, self.colors, str(self.name) + "+" + str(core1.name))

    def compute_or(self, core1):
        core0 = self.extend_colors(core1)
        core1 = core1.extend_colors(self)

        summed = core0.sum_with(core1)
        truth_positions = np.argwhere(summed.values > 0)
        orValues = np.zeros(shape=summed.values.shape)
        for pos in truth_positions:
            orValues[tuple(pos)] = 1
        return CoordinateCore(orValues, summed.colors, [self.name, "and", core1.name])

    def extend_colors(self, core1):
        core0 = self.clone()

        colorDict = {core0.colors[i]: alphabet[i] for i in range(len(core0.colors))}
        i = len(core0.colors)
        for color in core1.colors:
            if color not in colorDict:
                colorDict[color] = alphabet[i]
                i += 1

        added_colors = []
        added_shape = []
        for i, color in enumerate(core1.colors):
            if not color in core0.colors:
                added_colors.append(color)
                added_shape.append(core1.values.shape[i])

        core0_string = "".join([colorDict[color] for color in core0.colors])
        added_string = "".join([colorDict[color] for color in added_colors])

        premise_string = ",".join([core0_string, added_string])
        head_string = core0_string + added_string
        contraction_string = "->".join([premise_string, head_string])
        newValues = np.einsum(contraction_string, core0.values, np.ones(shape=added_shape))

        return CoordinateCore(newValues, core0.colors + added_colors, name=str(core0.name) + "_extended")

    def contract_common_colors(self, in_core1, exceptions=[]):
        core0 = self.clone()
        core1 = in_core1.clone()

        colorDict = {core0.colors[i]: alphabet[i] for i in range(len(core0.colors))}
        i = len(core0.colors)
        for color in core1.colors:
            if color not in colorDict:
                colorDict[color] = alphabet[i]
                i += 1
            elif color in exceptions:
                cor_color = color + "_cor"
                colorDict[cor_color] = alphabet[i]
                i += 1

        core0_string = "".join([colorDict[color] for color in core0.colors])
        core1_string = ""
        new_string = ""
        new_colors = []
        for color in core1.colors:
            if color in exceptions:
                core1_string = core1_string + colorDict[color + "_cor"]
            else:
                core1_string = core1_string + colorDict[color]

            if color in exceptions:
                new_string = new_string + colorDict[color] + colorDict[color + "_cor"]
                new_colors.append(color)
                new_colors.append(color + "_cor")
            elif color not in core0.colors:
                new_string = new_string + colorDict[color]
                new_colors.append(color)

        for color in core0.colors:
            if color not in core1.colors:
                new_string = new_string + colorDict[color]
                new_colors.append(color)

        premise_string = ",".join([core0_string, core1_string])
        contraction_string = "->".join([premise_string, new_string])

        newValues = np.einsum(contraction_string, core0.values, core1.values)

        return CoordinateCore(newValues, new_colors, [core0.name, "contracted", core1.name])

    def create_constant(self, variableColors, zero=False):
        constantShape = []
        constantColors = []
        for i, color in enumerate(self.colors):
            if color not in variableColors:
                constantShape.append(self.values.shape[i])
                constantColors.append(color)
        if zero:
            return CoordinateCore(np.zeros(constantShape), constantColors, "Constant")
        else:
            return CoordinateCore(np.ones(constantShape), constantColors, "Constant")

    def reduce_color(self, conColor):
        ## To be implemented for efficiency increase: Alternative using np.sum
        if conColor not in self.colors:
            raise ValueError("Color {} not found in core {} for reduction.".format(conColor, self.name))
        colorDict = {col: alphabet[i] for i, col in enumerate(self.colors)}
        contractionstring = "".join([colorDict[col] for col in self.colors]) + "," + colorDict[
            conColor] + "->" + "".join(([colorDict[col] for col in self.colors if col != conColor]))
        onesValues = np.ones(self.values.shape[self.colors.index(conColor)])
        newValues = np.einsum(contractionstring, self.values, onesValues)
        return CoordinateCore(newValues, [col for col in self.colors if col != conColor])

    ## For usage in model (former BasisCalculus methods)
    def weighted_exponentiation(self, weight):
        return CoordinateCore(np.exp(weight * self.values), self.colors)

    def count_satisfaction(self):
        return np.sum(self.values) / np.prod(self.values.shape)

    def normalize(self, sum=1):
        return CoordinateCore(self.values * (sum / np.sum(self.values)), self.colors, self.name)

    def exponentiate(self):
        return CoordinateCore(np.exp(self.values), self.colors, self.name)

    def logarithm(self):
        return CoordinateCore(np.log(self.values), self.colors, self.name)


## Small Test Skript
if __name__ == "__main__":
    core0_values = np.random.normal(size=(100, 10, 5))
    core0_colors = ["x1", "y2", "z3"]

    core0 = CoordinateCore(core0_values, core0_colors, "Sledz")
    # core0.reorder_colors(["b", "c", "a"])

    core1_values = np.random.normal(size=(10, 100, 5))
    core1_colors = ["y2", "x1", "z3"]

    core0.negate()
    core1 = CoordinateCore(core1_values, core1_colors, "Jaszczur")

    einSum_Core = core0.compute_and(core1)
    compare_Core = core0.compute_and(core1).negate(ignore_ones=True)

    toBeZero = einSum_Core.sum_with(compare_Core)

    print(np.linalg.norm(einSum_Core.values))
    print(np.linalg.norm(compare_Core.values))
    print(np.linalg.norm(toBeZero.values))
    print(einSum_Core.colors)
    print(compare_Core.colors)
    core0.sum_with(core1)



import tnreason.representation.ttl_to_csv as ttoc
import tnreason.representation.csv_to_ttl as ctot
import tnreason.representation.sampledf_to_factdf as stof
import tnreason.representation.sampledf_to_pairdf as stop

import tnreason.logic.sparse_cc as scc

import rdflib
import pandas as pd
import numpy as np
import scipy.sparse as scs

class SparseCoreGenerator:
    def __init__(self):
        self.factDf = None
        self.pairDf = None

    def load_ttl(self,ttlPath):
        if self.factDf is None:
            self.factDf = ttoc.generate_csv_list(ttlPath)
        else:
            self.factDf = ttoc.extend_csv_list(self.factDf, ttlPath)

    def load_factDf(self, factDf):
        if self.factDf is None:
            self.factDf = factDf
        else:
            self.factDf = pd.concatenate(self.factDf,factDf)

    def calculate_pairDf(self,extractionQuery):
        ## Need to implement extractionQuery - via ctot and rdflib?
        pass

    def load_sampleDf(self,sampleDf):
        rows = []
        cols = []
        for i, row in sampleDf.iterrows():
            for j, col in enumerate(sampleDf.columns):
                if row[col] == 1:
                    rows.append(i)
                    cols.append(j)
        self.repMatrix = scs.coo_matrix((np.ones(len(rows)), (rows, cols)), shape=(sampleDf.shape[0],sampleDf.shape[1]))

        self.dataIndex = sampleDf.index
        self.axiomIndex = list(sampleDf.columns)

    def generate_sparseAxiomCore(self,axiomList=None,axiomColor="var"):
        if axiomList == None:
            return scc.SparseCoordinateCore(core_values=self.repMatrix, core_colors=["j",axiomColor])
        else:
            axiomPositions = [self.axiomIndex.index(axiom) for axiom in axiomList]

            rows = []
            cols = []

            for i,j,v in zip(self.repMatrix.row, self.repMatrix.col, self.repMatrix.data):
                if j in axiomPositions:
                    rows.append(i)
                    cols.append(j)

            return scc.SparseCoordinateCore(core_values=scs.coo_matrix((np.ones(len(rows)),(rows, cols)), shape=self.repMatrix.shape),
                                            core_colors=["j", axiomColor])


if __name__ == "__main__":
    loader = SparseCoreGenerator()
    sampleDf = pd.read_csv("./demonstration/generation/synthetic_test_data/generated_sampleDf.csv")
    loader.load_sampleDf(sampleDf)

    sparsecore = loader.generate_sparseAxiomCore(["Moebel(z)","Bautischlerei(y)"])
    print(sparsecore.values)



def calculate_core_dict(variableCoresDict, fixedCoresDict, legKey):
    core_dict = {}
    for coreKey in fixedCoresDict:
        if coreKey != legKey:
            core_dict[coreKey] = fixedCoresDict[coreKey].clone().contract_common_colors(
                variableCoresDict[coreKey].clone())
        else:
            core_dict[coreKey] = fixedCoresDict[coreKey].clone()
    return core_dict

def affected_colors(coreColors, variableColors):
    for color in coreColors:
        if color in variableColors:
            return True
    return False

def color_equivalence(core1,core2):
    for color in core1.colors:
        if color not in core2.colors:
            return False
    for color in core2.colors:
        if color not in core1.colors:
            return False
    return True

def sum_coreList(coreList):
    resultList = []
    while len(coreList)>0:
        core1 = coreList.pop(0)
        for core2 in coreList.copy():
            if color_equivalence(core1,core2):
                core1 = core1.sum_with(core2)
                coreList.pop(coreList.index(core2))
        resultList.append(core1)
    return resultList

def calculate_coreList(atom_dict,expression,variableColors):
    if type(expression) == str:
        return [atom_dict[expression]]
    elif expression[0] == "not":
        preCoreList = calculate_coreList(atom_dict, expression[1], variableColors)
        coreList = []
        for core in preCoreList:
            coreList.append(core.negate(ignore_ones = True))
        coreList.append(preCoreList[0].create_constant(variableColors))
        return sum_coreList(coreList)
    elif expression[1] == "and":
        toBeSummed = []
        preCoreList0 = calculate_coreList(atom_dict, expression[0], variableColors)
        preCoreList2 = calculate_coreList(atom_dict, expression[2], variableColors)
        for core0 in preCoreList0:
            for core2 in preCoreList2:
                toBeSummed.append(core0.compute_and(core2))
        return sum_coreList(toBeSummed)
    else:
        raise ValueError("Expression {} not understood.".format(expression))

def operator_constant_from_coreList(coreList, variableColors):
    assert len(coreList) <= 2, "CoreList too long!"
    if len(coreList) == 1:
        if affected_colors(coreList[0].colors,variableColors):
            return coreList[0], coreList[0].create_constant(variableColors, zero=True)
        else:
            raise ValueError("Variable Core is disconnected!")
    else:
        if affected_colors(coreList[0].colors,variableColors):
            return coreList[0], coreList[1]
        else:
            return coreList[1], coreList[0]

## To support the older tests
def calculate_operator_and_constant(atom_dict, expression, variableKey, variableColors):
    return operator_constant_from_coreList(calculate_coreList(atom_dict,expression,variableColors),variableColors)


from tnreason.logic import coordinate_calculus as cc
from tnreason.logic import basis_calculus as bc

import numpy as np


def generate_rawCoreDict(expressionDict):
    ## ExpressionDict of structure key: expression
    rawCoreDict = {}
    for formulaKey in expressionDict:
        expression = expressionDict[formulaKey]
        rawCoreDict = {**rawCoreDict, **generate_factor_dict(expression, formulaKey=formulaKey, headType="empty")}
    return rawCoreDict


def generate_formulaCoreDict(expressionDict):
    ## ExpressionDict of structure key: [expression, weight]
    rawCoreDict = {}
    for formulaKey in expressionDict:
        expression = expressionDict[formulaKey][0]
        weight = expressionDict[formulaKey][1]
        rawCoreDict = {**rawCoreDict,
                       **generate_factor_dict(expression, formulaKey=formulaKey, weight=weight, headType="expFactor")}
    return rawCoreDict


def generate_exponentiationHeadValues(weight, headcolors, differentiated=False):
    if differentiated:
        values = np.zeros(shape=2)
    else:
        values = np.ones(shape=2)
    values[1] = np.exp(weight)
    return cc.CoordinateCore(values, headcolors)


def generate_factor_dict(expression, formulaKey="f0", weight=0, headType="truthEvaluation"):
    factorDict = create_formulaProcedure(expression, formulaKey)
    headColors = [formulaKey + "_" + str(expression)]
    if headType == "truthEvaluation":
        factorDict[formulaKey + "_" + str(expression) + "_" + headType] = cc.CoordinateCore(
            bc.create_truth_vec(), headColors)
    elif headType == "expFactor":
        factorDict[formulaKey + "_" + str(expression) + "_" + headType] = generate_exponentiationHeadValues(weight,
                                                                                                            headColors,
                                                                                                            differentiated=False)
    elif headType == "diffExpFactor":
        factorDict[formulaKey + "_" + str(expression) + "_" + headType] = generate_exponentiationHeadValues(weight,
                                                                                                            headColors,
                                                                                                            differentiated=True)
    elif headType == "empty":
        pass
    else:
        raise ValueError("Head Type {} not understood!".format(headType))
    return factorDict


## For the generation of Basis Calculus Instructions
def create_formulaProcedure(expression, formulaKey):
    addCoreKey = str(formulaKey) + "_" + str(expression) + "_subCore"
    if type(expression) == str:
        return {addCoreKey: cc.CoordinateCore(np.eye(2), [expression, formulaKey + "_" + expression], expression)}
    elif expression[0] == "not":
        if type(expression[1]) == str:
            return {addCoreKey: cc.CoordinateCore(bc.create_negation_tensor(),
                                                  [expression[1], formulaKey + "_" + str(expression)],
                                                  expression)}
        else:
            partsDict = create_formulaProcedure(expression[1], formulaKey)
            partsDict[addCoreKey] = cc.CoordinateCore(bc.create_negation_tensor(),
                                                      [formulaKey + "_" + str(expression[1]),
                                                       formulaKey + "_" + str(expression)], expression)
            return partsDict
    elif expression[1] == "and":
        if type(expression[0]) == str:
            partsDict0 = {}
            leftColor = expression[0]
        else:
            partsDict0 = create_formulaProcedure(expression[0], formulaKey)
            leftColor = formulaKey + "_" + str(expression[0])

        if type(expression[2]) == str:
            partsDict2 = {}
            rightColor = expression[2]
        else:
            partsDict2 = create_formulaProcedure(expression[2], formulaKey)
            rightColor = formulaKey + "_" + str(expression[2])

        return {**partsDict0, **partsDict2,
                addCoreKey: cc.CoordinateCore(bc.create_and_tensor(),
                                              [leftColor, rightColor, formulaKey + "_" + str(expression)],
                                              str(expression))}
        ## OLD: Resolving key conflicts in the dictionary
        # NOT NEEDED! If colliding, then same binary core

        ## Renaming cores of the right hand side to avoid key collision
        # partsDict2 = {}
        # for key in prePartsDict2:
        #    if key in partsDict0:
        #        partsDict2[key + "0"] = prePartsDict2[key]
        #    else:
        #        partsDict2[key + "0"] = prePartsDict2[key]

        ## Renaming colors of the right hand side to avoid duplicates (except for atoms)
        # colors0 = get_colors_from_coreDict(partsDict0)
        # preColors2 = get_colors_from_coreDict(partsDict2)
        # replaceColorDict = create_newColorDict(colors0, preColors2)
        # partsDict2 = replace_colors_in_coreDict(partsDict2, replaceColorDict)
        # if rightColor in replaceColorDict:
        #    rightColor = replaceColorDict[rightColor]


# def get_colors_from_coreDict(coreDict):
#     colors = []
#     for coreKey in coreDict:
#         for color in coreDict[coreKey].colors:
#             if color not in colors:
#                 colors.append(color)
#     return colors
#
#
# def create_newColorDict(colorsLeft, colorsRight):
#     newColorDict = {}
#     for color in colorsRight:
#         if color in colorsLeft and "_h" in color:
#             newColor = color
#             while newColor in colorsLeft:
#                 newColor = newColor + "0"
#             newColorDict[color] = newColor
#     return newColorDict
#
#
# def replace_colors_in_coreDict(coreDict, newColorDict):
#     for coreKey in coreDict:
#         for i, color in enumerate(coreDict[coreKey].colors):
#             if color in newColorDict:
#                 coreDict[coreKey].colors[i] = newColorDict[color]
#     return coreDict


if __name__ == "__main__":
    expression = [["not", "sledz"], "and", ["not", "sledz"]]

    solDict = create_formulaProcedure(expression, "Sledz")
    print([solDict[coreKey].colors for coreKey in solDict])
